"""
functions adapted from https://github.com/labsyspharm/ashlar to provide support for multi-threaded execution.
"""

import copy as copy
import sys

import numpy as np
import sklearn.linear_model
from alphabase.io.tempmmap import mmap_array_from_path
from tqdm.auto import tqdm

# import graph algorithms
try:
    from graph_tool import Graph as gtGraph
    from graph_tool import GraphView
    from graph_tool.generation import remove_parallel_edges
    from graph_tool.search import bfs_iterator
    from graph_tool.topology import label_components, shortest_path

    flavor = "graph-tool"
except ImportError:
    flavor = "networkx"

from ashlar import utils as utils
from ashlar.reg import EdgeAligner, Mosaic, warn_data

from scportrait.tools.stitch._utils.graphs import get_center_nodes, nx2gt
from scportrait.tools.stitch._utils.parallelilzation import execute_indexed_parallel, execute_parallel


class ParallelEdgeAligner(EdgeAligner):
    def __init__(
        self,
        reader,
        n_threads=20,
        channel=0,
        max_shift=15,
        alpha=0.01,
        max_error=None,
        randomize=False,
        filter_sigma=0.0,
        do_make_thumbnail=True,
        verbose=False,
    ):
        super().__init__(
            reader=reader,
            channel=channel,
            max_shift=max_shift,
            alpha=alpha,
            max_error=max_error,
            randomize=randomize,
            filter_sigma=filter_sigma,
            do_make_thumbnail=do_make_thumbnail,
            verbose=verbose,
        )

        self.n_threads = n_threads

        # set flavor to graph-tool if available
        try:
            from graph_tool import Graph as gtGraph

            self.flavor = "graph-tool"
        except ImportError:
            Warning(
                "graph-tool not available, using networkx as default. \n For stitching large datasets, graph-tool is recommended as it provides better performance."
            )
            self.flavor = "networkx"

    def compute_threshold(self):
        # Compute error threshold for rejecting aligments. We generate a
        # distribution of error scores for many known non-overlapping image
        # regions and take a certain percentile as the maximum allowable error.
        # The percentile becomes our accepted false-positive ratio.
        edges = self.neighbors_graph.edges
        num_tiles = self.metadata.num_images
        # If not enough tiles overlap to matter, skip this whole thing.
        if len(edges) <= 1:
            self.errors_negative_sampled = np.empty(0)
            self.max_error = np.inf
            return
        widths = np.array([self.intersection(t1, t2).shape.min() for t1, t2 in edges])
        w = widths.max()
        max_offset = self.metadata.size[0] - w
        # Number of possible pairs minus number of actual neighbor pairs.
        num_distant_pairs = num_tiles * (num_tiles - 1) // 2 - len(edges)
        # Reduce permutation count for small datasets -- there are fewer
        # possible truly distinct strips with fewer tiles. The calculation here
        # is just a heuristic, not rigorously derived.
        n = 1000 if num_distant_pairs > 8 else (num_distant_pairs + 1) * 10
        pairs = np.empty((n, 2), dtype=int)
        offsets = np.empty((n, 2), dtype=int)
        # Generate n random non-overlapping image strips. Strips are always
        # horizontal, across the entire image width.
        max_tries = 100
        if self.randomize is False:
            random_state = np.random.RandomState(0)
        else:
            random_state = np.random.RandomState()
        for i in range(n):
            # Limit tries to avoid infinite loop in pathological cases.
            for _current_try in range(max_tries):
                t1, t2 = random_state.randint(self.metadata.num_images, size=2)
                o1, o2 = random_state.randint(max_offset, size=2)
                # Check for non-overlapping strips and abort the retry loop.
                if t1 != t2 and (t1, t2) not in edges:
                    # Different, non-neighboring tiles -- always OK.
                    break
                elif t1 == t2 and abs(o1 - o2) > w:
                    # Same tile OK if strips don't overlap within the image.
                    break
                elif (t1, t2) in edges:
                    # Neighbors OK if either strip is entirely outside the
                    # expected overlap region (based on nominal positions).
                    its = self.intersection(t1, t2, np.repeat(w, 2))
                    ioff1, ioff2 = its.offsets[:, 0]
                    if (
                        its.shape[0] > its.shape[1]
                        or o1 < ioff1 - w
                        or o1 > ioff1 + w
                        or o2 < ioff2 - w
                        or o2 > ioff2 + w
                    ):
                        break
            else:
                # Retries exhausted. This should be very rare.
                warn_data("Could not find non-overlapping strips in {max_tries} tries")
            pairs[i] = t1, t2
            offsets[i] = o1, o2

        def register(t1, t2, offset1, offset2):
            img1 = self.reader.read(t1, self.channel)[offset1 : offset1 + w, :]
            img2 = self.reader.read(t2, self.channel)[offset2 : offset2 + w, :]
            _, error = utils.register(img1, img2, self.filter_sigma, upsample=1)
            return error

        # prepare arguments for executor
        args = []
        for (t1, t2), (offset1, offset2) in zip(pairs, offsets, strict=False):
            arg = (t1, t2, offset1, offset2)
            args.append(copy.deepcopy(arg))

        errors = execute_indexed_parallel(
            register,
            args=args,
            tqdm_args={
                "file": sys.stdout,
                "disable": not self.verbose,
                "desc": "    quantifying alignment error",
            },
            n_threads=self.n_threads,
        )

        errors = np.array(errors)
        self.errors_negative_sampled = errors
        self.max_error = np.percentile(errors, self.alpha * 100)

    def register_all(self):
        args = []
        for t1, t2 in self.neighbors_graph.edges:
            arg = (t1, t2)
            args.append(copy.deepcopy(arg))

        execute_parallel(
            self.register_pair,
            args=args,
            tqdm_args={
                "file": sys.stdout,
                "disable": not self.verbose,
                "desc": "                  aligning edge",
            },
            n_threads=self.n_threads,
        )

        # need to reorder the errors to match the order of the edges
        self.all_errors = np.array([x[1] for x in self._cache.values()])

        # Set error values above the threshold to infinity.
        for k, v in self._cache.items():
            if v[1] > self.max_error or np.any(np.abs(v[0]) > self.max_shift_pixels):
                self._cache[k] = (v[0], np.inf)

        self.cached_errors = self._cache.copy()  # save as a backup

    def _build_spanning_tree_gt(self):
        from networkx import Graph as nxGraph

        g = nxGraph()
        g.add_nodes_from(self.neighbors_graph)
        g.add_weighted_edges_from(
            (t1, t2, error) for (t1, t2), (_, error) in self.cached_errors.items() if np.isfinite(error)
        )

        # convert to a graph-tool graph
        gtG = nx2gt(g)

        spanning_tree = gtGraph(gtG)
        spanning_tree.clear_edges()

        # label the components in a property map
        c = label_components(gtG)[0]
        components = np.unique(c.a)

        centers = []
        for i in components:
            u = GraphView(gtG, vfilt=c.a == i)

            center = get_center_nodes(u)
            centers.append(center)
            vertices = list(u.vertices())
            for vertix in vertices:
                vlist, elist = shortest_path(u, center, vertix, weights=u.ep.weight)
                spanning_tree.add_edge_list(elist)

        remove_parallel_edges(spanning_tree)

        self.spanning_tree = spanning_tree
        self.centers_spanning_tree = centers

    def _build_spanning_tree_nxg(self):
        import networkx as nx
        from networkx import Graph as nxGraph

        g = nxGraph()
        g.add_nodes_from(self.neighbors_graph)
        g.add_weighted_edges_from((t1, t2, error) for (t1, t2), (_, error) in self._cache.items() if np.isfinite(error))
        spanning_tree = nxGraph()
        spanning_tree.add_nodes_from(g)

        centers = []

        for c in nx.connected_components(g):
            cc = g.subgraph(c)
            center = nx.center(cc)[0]
            centers.append(center)
            paths = nx.single_source_dijkstra_path(cc, center).values()
            for path in paths:
                nx.add_path(spanning_tree, path)

        self.spanning_tree = spanning_tree
        self.centers_spanning_tree = centers

    def build_spanning_tree(self):
        if self.flavor == "graph-tool":
            print("using graph-tool to build spanning tree")
            self._build_spanning_tree_gt()
        if self.flavor == "networkx":
            print("using networkx to build spanning tree")
            self._build_spanning_tree_nxg()

    def _calculate_positions_gt(self):
        shifts = {}
        _components = []

        # label the components in a property map
        c = label_components(self.spanning_tree)[0]
        components = np.unique(c.a)

        for ix, i in enumerate(components):
            u = GraphView(self.spanning_tree, vfilt=c.a == i)
            nodes = list(u.get_vertices())
            _components.append(set(nodes))

            center = self.centers_spanning_tree[ix]

            shifts[center] = np.array([0, 0])

            if len(nodes) > 1:
                for edge in bfs_iterator(u, source=center):
                    source, dest = edge
                    source = int(source)
                    dest = int(dest)
                    if source not in shifts:
                        source, dest = dest, source
                    shift = self.register_pair(source, dest)[0]
                    shifts[dest] = shifts[source] + shift

        if shifts:
            self.shifts = np.array([s for _, s in sorted(shifts.items())])
            self.positions = self.metadata.positions + self.shifts
            self.components_spanning_tree = _components
        else:
            # TODO: fill in shifts and positions with 0x2 arrays
            raise NotImplementedError("No images")

    def _calculate_positions_nxg(self):
        import networkx as nx

        shifts = {}
        for c in nx.connected_components(self.spanning_tree):
            cc = self.spanning_tree.subgraph(c)
            center = nx.center(cc)[0]
            shifts[center] = np.array([0, 0])
            for edge in nx.traversal.bfs_edges(cc, center):
                source, dest = edge
                if source not in shifts:
                    source, dest = dest, source
                shift = self.register_pair(source, dest)[0]
                shifts[dest] = shifts[source] + shift
        if shifts:
            self.shifts = np.array([s for _, s in sorted(shifts.items())])
            self.positions = self.metadata.positions + self.shifts
            self.components_spanning_tree = nx.connected_components(self.spanning_tree)
        else:
            # TODO: fill in shifts and positions with 0x2 arrays
            raise NotImplementedError("No images")

    def calculate_positions(self):
        if self.flavor == "graph-tool":
            self._calculate_positions_gt()
        if self.flavor == "networkx":
            self._calculate_positions_nxg()

    def fit_model(self):
        components = self.components_spanning_tree
        components = sorted(components, key=len, reverse=True)

        # Fit LR model on positions of largest connected component.
        cc0 = list(components[0])
        self.lr = sklearn.linear_model.LinearRegression()
        self.lr.fit(self.metadata.positions[cc0], self.positions[cc0])
        # Fix up degenerate transform matrix. This happens when the spanning
        # tree is completely edgeless or cc0's metadata positions fall in a
        # straight line. In this case we fall back to the identity transform.
        if np.linalg.det(self.lr.coef_) < 1e-3:
            # FIXME We should probably exit here, not just warn. We may provide
            # an option to force it anyway.
            warn_data("Could not align enough edges, proceeding anyway with original" " stage positions.")
            self.lr.coef_ = np.diag(np.ones(2))
            self.lr.intercept_ = np.zeros(2)

        # Adjust position of remaining components so their centroids match
        # the predictions of the model.
        for cc in components[1:]:
            nodes = list(cc)
            centroid_m = np.mean(self.metadata.positions[nodes], axis=0)
            centroid_f = np.mean(self.positions[nodes], axis=0)
            shift = self.lr.predict([centroid_m])[0] - centroid_f
            self.positions[nodes] += shift
        # Adjust positions and model intercept to put origin at 0,0.
        self.origin = self.positions.min(axis=0)
        self.positions -= self.origin
        self.lr.intercept_ -= self.origin
        self.centers = self.positions + self.metadata.size / 2


class ParallelMosaic(Mosaic):
    def __init__(
        self,
        aligner,
        shape,
        n_threads=20,
        channels=None,
        ffp_path=None,
        dfp_path=None,
        flip_mosaic_x=False,
        flip_mosaic_y=False,
        barrel_correction=None,
        verbose=False,
    ):
        super().__init__(
            aligner=aligner,
            shape=shape,
            channels=channels,
            ffp_path=ffp_path,
            dfp_path=dfp_path,
            flip_mosaic_x=flip_mosaic_x,
            flip_mosaic_y=flip_mosaic_y,
            barrel_correction=barrel_correction,
            verbose=verbose,
        )

        self.n_threads = n_threads

    def assemble_channel_parallel(
        self,
        channel,
        ch_index: int,
        out: np.ndarray | None = None,
        hdf5_path: str | None = None,
        tqdm_args=None,
        n_percent: int = 10,
    ):
        """This function assembles a single channel of the mosaic writing to the same HDF5 file being used as a mmap array in the backend."""
        if out is None:
            if hdf5_path is not None:
                out = mmap_array_from_path(hdf5_path)
            else:
                out = np.zeros(self.shape, self.dtype)
        else:
            if out.shape != self.shape:
                raise ValueError(f"out array shape {out.shape} does not match Mosaic shape {self.shape}")
            if hdf5_path is None:
                raise ValueError(
                    "if specifying an out array, you also need to pass the HDF5 path of the memory mapped temparray"
                )

        # Set up tqdm arguments if not provided
        if tqdm_args is None:
            tqdm_args = {
                "file": sys.stdout,
                "disable": not self.verbose,
                "desc": f"assembling channel {ch_index}",
                "total": len(self.aligner.positions),
            }

        total_positions = len(self.aligner.positions)
        update_interval = int(total_positions * (n_percent / 100))
        last_update = 0

        # Assemble channel with progress updates every n%
        with tqdm(**tqdm_args) as pbar:
            for si, position in enumerate(self.aligner.positions):
                img = self.aligner.reader.read(c=channel, series=si)
                img = self.correct_illumination(img, channel)
                utils.paste(out[ch_index, :, :], img, position, func=utils.pastefunc_blend)

                # Update progress bar every n%
                if si >= last_update + update_interval:
                    pbar.update(update_interval)
                    last_update = si

            # Ensure the progress bar reaches 100% at the end
            if pbar.n < total_positions:
                pbar.update(total_positions - pbar.n)

        # Memory-conserving axis flips
        if self.flip_mosaic_x:
            for i in range(len(out)):
                out[i] = out[i, ::-1]
        if self.flip_mosaic_y:
            for i in range(len(out) // 2):
                out[[i, -i - 1]] = out[[-i - 1, i]]

        return None



from pathlib import Path
from typing import Literal

import h5py
from anndata import AnnData
from anndata._io.h5ad import _clean_uns, _read_raw, read_dataframe, read_elem

from scportrait.pipeline._utils.constants import DEFAULT_NAME_SINGLE_CELL_IMAGES, IMAGE_DATACONTAINER_NAME


def read_h5sc(filename: str | Path) -> AnnData:
    """Read scportrait's single-cell image dataset format.

    Args:
        filename: Path to the file to read.
        mode: Mode in which to open the file.

    Returns:
        An AnnData object with obsm["single_cell_images"] containing a memory-backed array of the single-cell images.
    """
    mode = "r"  # hard code as there is no benefit from letting the user set this manually as writing back and changing values is currently not supported
    d = {}

    # connect to h5py file
    f = h5py.File(filename, mode)

    attributes = ["varm", "obsp", "varp", "uns", "layers"]
    df_attributes = ["obs", "var"]

    if "encoding-type" in f.attrs:
        attributes.extend(df_attributes)
    else:
        for k in df_attributes:
            if k in f:  # Backwards compat
                d[k] = read_dataframe(f[k])

    d.update({k: read_elem(f[k]) for k in attributes if k in f})

    d["raw"] = _read_raw(f, attrs={"var", "varm"})
    adata = AnnData(**d)

    # Backwards compat to <0.7
    if isinstance(f["obs"], h5py.Dataset):
        _clean_uns(adata)

    adata.obsm[DEFAULT_NAME_SINGLE_CELL_IMAGES] = f.get(IMAGE_DATACONTAINER_NAME)
    return adata



"""
h5sc
=======

Functions to work with scPortrait's standardized single-cell data format.
"""

from .operations import get_image_index, get_image_with_cellid

__all__ = ["get_image_with_cellid", "get_image_index"]



from pathlib import Path

from anndata import AnnData

from scportrait.data._datasets import _get_remote_dataset
from scportrait.io.h5sc import read_h5sc


def dataset2_h5sc() -> AnnData:
    """Get example single-cell image dataset derived from dataset 2.

    Returns:
        Loaded h5sc dataset.
    """
    DATASET = "dataset2_h5sc"
    URL = "https://zenodo.org/records/15164629/files/single_cells.h5sc?download=1"
    NAME = "single_cells.h5sc"
    path = _get_remote_dataset(DATASET, URL, NAME, archive_format=None, outfile_name=NAME)
    return read_h5sc(path)



import warnings
from collections.abc import Iterable
from numbers import Integral

import matplotlib as mpl
import matplotlib.pyplot as plt
import spatialdata
import xarray
from matplotlib.axes import Axes

PALETTE = [
    "blue",
    "green",
    "red",
    "yellow",
    "purple",
    "orange",
    "pink",
    "cyan",
    "magenta",
    "lime",
    "teal",
    "lavender",
    "brown",
    "beige",
    "maroon",
    "mint",
    "olive",
    "apricot",
    "navy",
    "grey",
    "white",
    "black",
]

from scportrait.pipeline._utils.helper import _check_for_spatialdata_plot


def _get_shape_element(sdata, element_name) -> tuple[int, int]:
    """Get the x, y shape of the element in the spatialdata object.

    Args:
        sdata: SpatialData object containing the element.
        element_name: Name of the element to get the shape of.

    Returns:
        Tuple containing the shape of the x, y coordinates of the element.
    """
    if isinstance(sdata[element_name], xarray.DataTree):
        shape = sdata[element_name].scale0.image.shape
    else:
        shape = sdata[element_name].data.shape

    if len(shape) == 3:
        _, x, y = shape
    elif len(shape) == 2:
        x, y = shape
    return x, y


def _create_figure_dpi(x: float, y: float, dpi: int | None = 300) -> tuple[plt.Figure, Axes]:
    """Helper function to create a figure to plot a given image with x, y resolution at the specified DPI.
    Args:
        x: x resolution of the image.
        y: y resolution of the image.
        dpi: Dots per inch of the image. Defaults to 300.
    Returns:
        Tuple containing the figure and axis objects.
    """
    # create figure and axis if it does not exist
    if dpi is None:
        dpi = 300

    axs_length = x / dpi
    axs_width = y / dpi

    fig = plt.figure(figsize=(axs_width, axs_length), frameon=False, dpi=dpi)
    ax = fig.add_axes([0.0, 0.0, 1.0, 1.0])
    return fig, ax


def plot_image(
    sdata: spatialdata.SpatialData,
    image_name: str,
    channel_names: list[str] | list[int] | None = None,
    palette: list[str] | None = None,
    cmap: mpl.colors.ListedColormap | None = None,
    title: str | None = None,
    title_fontsize: int = 20,
    dpi: int | None = None,
    norm=None,
    ax: Axes = None,
    return_fig: bool = False,
    show_fig: bool = True,
) -> plt.Figure | None:
    """Plot the image with the given name from the spatialdata object.

    Args:
        sdata: SpatialData object containing the image.
        image_name: Name of the image to plot.
        channel_names: List of channel names to plot. If None then all channels will be plotted. Defaults to None.
        palette: List of colors to use for the channels. If None then the default palette will be used. Defaults to None.
        title: Title of the plot. Defaults to None.
        title_fontsize: Font size of the title in points. Defaults to 20.
        dpi: Dots per inch of the plot. Defaults to None.
        ax: Matplotlib axis object to plot on. Defaults to None.
        return_fig: Whether to return the figure. Defaults to False.
        show_fig: Whether to show the figure. Defaults to True.

    Returns:
        Matplotlib figure object if return_fig is True.
    """
    _check_for_spatialdata_plot()

    if ax is not None:
        if dpi is not None:
            warnings.warn("DPI is ignored when an axis is provided.", stacklevel=2)
    else:
        # get size of spatialdata object to plot (required for calculating figure size if DPI is set)
        x, y = _get_shape_element(sdata, image_name)

        # create figure and axis if it does not exist
        fig, ax = _create_figure_dpi(x=x, y=y, dpi=dpi)

    if palette is None and cmap is None:
        if channel_names is None:
            palette = None
        else:
            palette = PALETTE[: len(channel_names)]

    # plot figure
    sdata.pl.render_images(image_name, channel=channel_names, palette=palette, cmap=cmap, norm=norm).pl.show(
        ax=ax, colorbar=False
    )
    ax.set_axis_off()
    ax.set_title(title, fontsize=title_fontsize)

    if return_fig:
        return fig
    elif show_fig:
        plt.show()
    return None


def plot_segmentation_mask(
    sdata: spatialdata.SpatialData,
    masks: list[str],
    background_image: str | None = "input_image",
    selected_channels: int | list[int] | None = None,
    max_channels_to_plot: int = 4,
    title: str | None = None,
    title_fontsize: int = 20,
    line_width: int = 1,
    line_color: str = "white",
    line_alpha: float = 1,
    fill_alpha: float = 0,
    fill_color: str | None = None,
    dpi: int | None = None,
    ax: plt.Axes | None = None,
    return_fig: bool = False,
    show_fig: bool = True,
) -> plt.Figure | None:
    """Visualize segmentation masks over selected background image.
    Will transform the provided labels layers to polygons and plot them over the background image.
    Requires an installed spatialdata_plot package.

    Args:
        sdata: SpatialData object containing the input image and segmentation mask.
        masks: List of keys identifying the segmentation masks to plot.
        background_image: Key identifying the background image to plot the segmentation masks on. Defaults to "input_image". If set to None then only the segmentation masks will be plotted as outlines.
        selected_channels: List of indices of the channels in the background image to plot. If None then the first `max_channels_to_plot` channels will be plotted. Defaults to None.
        title: Title of the plot. Defaults to None.
        title_fontsize: Font size of the title in points.
        line_width: Width of the outline of the segmentation masks.
        line_color: Color of the outline of the segmentation masks.
        line_alpha: Alpha value of the outline of the segmentation masks.
        fill_alpha: Alpha value of the fill of the segmentation masks.
        fill_color: Color of the fill of the segmentation masks. If None then no fill will be used. Defaults to None.
        dpi: Dots per inch of the plot. Defaults to None.
        axs: Matplotlib axis object to plot on. Defaults to None.
        return_fig: Whether to return the figure. Defaults to False.
        show_fig: Whether to show the figure. Defaults to True.
    """
    # check for spatialdata_plot
    _check_for_spatialdata_plot()

    # get relevant information from spatialdata object

    if ax is not None:
        if dpi is not None:
            warnings.warn("DPI is ignored when an axis is provided.", stacklevel=2)
    else:
        # get size of spatialdata object to plot (required for calculating figure size if DPI is set)
        x, y = _get_shape_element(sdata, masks[0])

        # create figure and axis if it does not exist
        fig, ax = _create_figure_dpi(x=x, y=y, dpi=dpi)

    # plot background image if desired
    if background_image is not None:
        # get channel names
        if background_image not in sdata:
            raise ValueError(f"Background image {background_image} not found in sdata object.")

        if isinstance(sdata[background_image], xarray.DataTree):
            channel_names = sdata[background_image].scale0.c.values
        else:
            channel_names = sdata[background_image].c.values

        # required work around because spatialdata plot only supports int and not things like int64
        if all(isinstance(x, Integral) for x in channel_names):
            channel_names = [int(x) for x in channel_names]

        c = len(channel_names)

        # do not plot more than `max_channels_to_plot` overlapping channels
        if selected_channels is not None:
            if not isinstance(selected_channels, Iterable):
                selected_channels = [selected_channels]
            channel_names = [channel_names[i] for i in selected_channels]
            c = len(channel_names)
            palette = [PALETTE[x] for x in selected_channels]
        else:
            if c > max_channels_to_plot:
                c = 4
            palette = PALETTE[:c]
            channel_names = list(channel_names[:c])

        sdata.pl.render_images(background_image, channel=channel_names, palette=palette).pl.show(ax=ax, colorbar=False)

    # plot selected segmentation masks
    for mask in masks:
        assert mask in sdata, f"Mask {mask} not found in sdata object."
        if f"{mask}_vectorized" not in sdata:
            sdata[f"{mask}_vectorized"] = spatialdata.to_polygons(sdata[mask])
        sdata.pl.render_shapes(
            f"{mask}_vectorized",
            color=fill_color,
            fill_alpha=fill_alpha,
            outline_alpha=line_alpha,
            outline_width=line_width,
            outline_color=line_color,
        ).pl.show(ax=ax)

    # configure axes
    ax.axis("off")
    ax.set_title(title, fontsize=title_fontsize)

    # return elements
    if return_fig:
        return fig
    elif show_fig:
        plt.show()
    return None


def plot_shapes(
    sdata: spatialdata.SpatialData,
    shapes_layer: str,
    title: str | None = None,
    title_fontsize: int = 20,
    fill_color: str = "grey",
    fill_alpha: float = 1,
    outline_color: str = "black",
    outline_alpha: float = 0,
    outline_width: float = 1,
    cmap: str = None,
    dpi: int | None = None,
    ax: plt.Axes | None = None,
    return_fig: bool = False,
    show_fig: bool = True,
) -> plt.Figure | None:
    """Visualize shapes layer.

    Args:
        sdata: SpatialData object containing the shapes layer.
        shapes_layer: Key identifying the shapes layer to plot.
        title: Title of the plot.
        title_fontsize: Font size of the title in points.
        fill_color: Color of the fill of the shapes.
        fill_alpha: Alpha value of the fill of the shapes.
        outline_color: Color of the outline of the shapes.
        outline_alpha: Alpha value of the outline of the shapes.
        outline_width: Width of the outline of the shapes.
        cmap: Colormap to use for the shapes.
        dpi: Dots per inch of the plot.
        ax: Matplotlib axis object to plot on.
        return_fig: Whether to return the figure.
        show_fig: Whether to show the figure.

    Returns:
        Matplotlib figure object if return_fig is True.
    """

    # check for spatialdata_plot
    _check_for_spatialdata_plot()

    if ax is not None:
        if dpi is not None:
            warnings.warn("DPI is ignored when an axis is provided.", stacklevel=2)
    else:
        # get size of spatialdata object to plot (required for calculating figure size if DPI is set)
        x, y = _get_shape_element(sdata, shapes_layer)

        # create figure and axis if it does not exist
        fig, ax = _create_figure_dpi(x=x, y=y, dpi=dpi)

    # plot selected shapes layer
    assert shapes_layer in sdata, f"Shapes layer {shapes_layer} not found in sdata object."

    sdata.pl.render_shapes(
        f"{shapes_layer}",
        fill_alpha=fill_alpha,
        color=fill_color,
        outline_alpha=outline_alpha,
        outline_color=outline_color,
        outline_width=outline_width,
        cmap=cmap,
    ).pl.show(ax=ax)

    ax.axis("off")
    ax.set_title(title, fontsize=title_fontsize)

    # return elements
    if return_fig:
        return fig
    elif show_fig:
        plt.show()
    return None


def plot_labels(
    sdata: spatialdata.SpatialData,
    label_layer: str,
    title: str | None = None,
    title_fontsize: int = 20,
    color: str = "grey",
    fill_alpha: float = 1,
    cmap: str = None,
    palette: dict | list = None,
    groups: list = None,
    norm: mpl.colors.Normalize = None,
    vectorized: bool = False,
    dpi: int | None = None,
    ax: plt.Axes | None = None,
    return_fig: bool = False,
    show_fig: bool = True,
) -> plt.Figure | None:
    """Plot the segmentation mask on the input image.

    Args:
        sdata: SpatialData object containing the input image and segmentation mask.
        labels_layer: Key identifying the label layer to plot.
        title: Title of the plot.
        title_fontsize: Font size of the title in points.
        color: color to plot the label layer in. Can be a string specifying a specific color or linking to a column in an annotating table.
        fill_alpha: Alpha value of the fill of the segmentation masks.
        cmap: Colormap to use for the labels.
        palette: Palette for discrete annotations. List of valid color names that should be used for the categories. Must match the number of groups. The list can contain multiple palettes (one per group) to be visualized. If groups is provided but not palette, palette is set to default “lightgray”.
        groups: When using color and the key represents discrete labels, groups can be used to show only a subset of them. Other values are set to NA.
        norm: Colormap normalization for continuous annotations
        vectorized: Whether to plot a vectorized version of the labels.
        dpi: Dots per inch of the plot.
        axs: Matplotlib axis object to plot on.
        return_fig: Whether to return the figure.
        show_fig: Whether to show the figure.

    Returns:
        Matplotlib figure object if return_fig is True.
    """
    # check for spatialdata_plot
    _check_for_spatialdata_plot()

    if ax is not None:
        if dpi is not None:
            warnings.warn("DPI is ignored when an axis is provided.", stacklevel=2)
    else:
        # get size of spatialdata object to plot (required for calculating figure size if DPI is set)
        x, y = _get_shape_element(sdata, label_layer)

        # create figure and axis if it does not exist
        fig, ax = _create_figure_dpi(x=x, y=y, dpi=dpi)

    # plot selected segmentation masks
    if vectorized:
        if f"{label_layer}_vectorized" not in sdata:
            sdata[f"{label_layer}_vectorized"] = spatialdata.to_polygons(sdata[label_layer])
        found_annotation = None
        for table in sdata.tables:
            annotated_regions = sdata.get_annotated_regions(sdata[table])
            for region in annotated_regions:
                if region == label_layer:
                    found_annotation = region
                    annotating_table = sdata[table].copy()
                    annotating_table.uns["spatialdata_attrs"]["region"] = f"{label_layer}_vectorized"
                    annotating_table.obs["region"] = f"{label_layer}_vectorized"
                    annotating_table.obs["region"] = annotating_table.obs["region"].astype("category")

                    # check for annotating column
                    if color in annotating_table.obs:
                        annotating_table.obs[color] = (
                            annotating_table.obs[color].astype("category").cat.remove_unused_categories()
                        )
                        # check for NaN values
                        if annotating_table.obs[color].isna().sum() > 0:
                            # NaN values need to be filled as otherwise the plotting will throw an error
                            if "NaN" not in annotating_table.obs[color].cat.categories:
                                annotating_table.obs[color] = annotating_table.obs[color].cat.add_categories("NaN")
                            annotating_table.obs[color] = annotating_table.obs[color].fillna("NaN")
                    annotating_table = spatialdata.models.TableModel.parse(annotating_table)
                    break
        if found_annotation is not None:
            sdata["_annotation"] = annotating_table
            sdata.pl.render_shapes(
                f"{label_layer}_vectorized",
                color=color,
                fill_alpha=fill_alpha,
                outline_alpha=0,
                cmap=cmap,
                palette=palette,
                groups=groups,
                norm=norm,
            ).pl.show(ax=ax)
            del sdata["_annotation"]  # delete element again after plotting
        else:
            try:
                sdata.pl.render_labels(
                    f"{label_layer}",
                    color=color,
                    fill_alpha=fill_alpha,
                    outline_alpha=1,
                    cmap=cmap,
                    palette=palette,
                    groups=groups,
                    norm=norm,
                ).pl.show(ax=ax)
            except Exception as err:
                raise Exception from err

    else:
        sdata.pl.render_labels(
            f"{label_layer}",
            color=color,
            fill_alpha=fill_alpha,
            cmap=cmap,
            palette=palette,
            groups=groups,
            norm=norm,
        ).pl.show(ax=ax)

    # configure axes
    ax.axis("off")
    ax.set_title(title, fontsize=title_fontsize)

    # return elements
    if return_fig:
        return fig
    elif show_fig:
        plt.show()
    return None



import multiprocessing as mp
import os
import shutil
import sys
import time
import timeit
import traceback
import warnings
from multiprocessing import current_process

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import xarray
from alphabase.io import tempmmap
from dask.array.core import Array as daskArray
from PIL import Image
from tqdm.auto import tqdm

from scportrait.io.daskmmap import dask_array_from_path
from scportrait.pipeline._base import ProcessingStep
from scportrait.pipeline._utils.constants import DEFAULT_CHANNELS_NAME, DEFAULT_FILTER_ADDTIONAL_FILE, DEFAULT_MASK_NAME
from scportrait.pipeline._utils.segmentation import _return_edge_labels, sc_any, shift_labels


class Segmentation(ProcessingStep):
    """Segmentation helper class used for creating segmentation workflows.

    Attributes:
        maps (dict(str)): Segmentation workflows based on the :class:`.Segmentation` class can use maps for saving and loading checkpoints and perform. Maps can be numpy arrays

        DEFAULT_FILTER_ADDTIONAL_FILE (str, default ``filtered_classes.csv``)
        PRINT_MAPS_ON_DEBUG (bool, default ``False``)

        identifier (int, default ``None``): Only set if called by :class:`ShardedSegmentation`. Unique index of the shard.
        window (list(tuple), default ``None``): Only set if called by :class:`ShardedSegmentation`. Defines the window which is assigned to the shard. The window will be applied to the input. The first element refers to the first dimension of the image and so on. For example use ``[(0,1000),(0,2000)]`` To crop the image to `1000 px height` and `2000 px width` from the top left corner.
        input_path (str, default ``None``): Only set if called by :class:`ShardedSegmentation`. Location of the input hdf5 file. During sharded segmentation the :class:`ShardedSegmentation` derived helper class will save the input image in form of a hdf5 file. This makes the input image available for parallel reading by the segmentation processes.

    Example:
        .. code-block:: python

            def process(self):
                # two maps are initialized
                self.maps = {"map0": None, "map1": None}

                # its checked if the segmentation directory already contains these maps and they are then loaded. The index of the first map which has not been found is returned. It indicates the step where computation needs to resume
                current_step = self.load_maps_from_disk()

                if current_step <= 0:
                    # do stuff and generate map0
                    self.save_map("map0")

                if current_step <= 1:
                    # do stuff and generate map1
                    self.save_map("map1")

    """

    # setup log and plotting behaviour
    CLEAN_LOG = True
    PRINT_MAPS_ON_DEBUG = True

    # load default values from constants.py
    DEFAULT_FILTER_ADDTIONAL_FILE = DEFAULT_FILTER_ADDTIONAL_FILE
    DEFAULT_CHANNELS_NAME = DEFAULT_CHANNELS_NAME
    DEFAULT_MASK_NAME = DEFAULT_MASK_NAME

    channel_colors = [
        "#0000FF",
        "#00FF00",
        "#FF0000",
        "#FFE464",
        "#9b19f5",
        "#ffa300",
        "#dc0ab4",
        "#b3d4ff",
        "#00bfa0",
    ]

    def __init__(
        self,
        config,
        directory,
        nuc_seg_name,
        cyto_seg_name,
        _tmp_image_path,
        project_location,
        debug,
        overwrite,
        project,
        filehandler,
        from_project: bool = False,
        **kwargs,
    ):
        super().__init__(
            config,
            directory,
            project_location,
            debug=debug,
            overwrite=overwrite,
            project=project,
            filehandler=filehandler,
            from_project=from_project,
        )

        if self.directory is not None:
            # only clean directory if a proper directoy is passed
            if self.CLEAN_LOG:
                self._clean_log_file()

        self._check_config()

        # if _tmp_seg is passed as an argument execute this following code (this only applies to some cases)
        if "_tmp_seg_path" in kwargs.keys():
            self._tmp_seg_path = kwargs["_tmp_seg_path"]

            # remove _tmp_seg from kwargs so that underlying classes do not need to account for it
            kwargs.pop("_tmp_seg_path")

        self.identifier = None
        self.window = None
        self.input_path = None
        self.is_shard = False

        # additional parameters to configure level of debugging for developers
        self.save_filter_results = False
        self.nuc_seg_name = nuc_seg_name
        self.cyto_seg_name = cyto_seg_name
        self._tmp_image_path = _tmp_image_path
        self.processes_per_GPU = None
        self.n_processes = 1

    def _check_config(self):
        """Check if the configuration is valid."""

        DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING = 0.95

        # optional config parameters that can be overridden through the config file
        if "chunk_size" in self.config.keys():
            self.chunk_size = self.config["chunk_size"]
        else:
            self.chunk_size = 50

        if "match_masks" in self.config.keys():
            self.match_masks = self.config["match_masks"]
        else:
            self.match_masks = True
            self.filtering_threshold_mask_matching = DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING

        if "filtering_threshold_mask_matching" in self.config.keys():
            self.filtering_threshold_mask_matching = self.config["filtering_threshold_mask_matching"]
        else:
            self.filtering_threshold_mask_matching = DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING

    def _check_gpu_status(self):
        if torch.cuda.is_available():
            self.use_GPU = True
            self.device = "cuda"
            self.nGPUs = torch.cuda.device_count()
        # check if MPS is available
        elif torch.backends.mps.is_available():
            self.use_GPU = True
            self.device = torch.device("mps")
            self.nGPUs = 1

        # default to CPU
        else:
            self.use_GPU = False
            self.device = torch.device("cpu")
            self.nGPUs = 0

    def _setup_processing(self):
        """
        Checks and updates the GPU status.
        """
        self._check_gpu_status()

        # compare with config configuration
        if "nGPUs" in self.config.keys():
            nGPUs = self.config["nGPUs"]
            if nGPUs == "max":
                self.log(f"Segmentation will be performed with all {self.nGPUs} found GPUs.")
            elif self.nGPUs != nGPUs:
                self.log(f"Found {self.nGPUs} available GPUS but {nGPUs} GPUs specified in config.")
                if self.nGPUs >= 1 and nGPUs >= 1:
                    self.nGPUs = min(self.nGPUs, nGPUs)
                    self.log(f"Will proceed with the number of GPUs specified in config ({self.nGPUs}).")
                else:
                    self.log(f"Segmentation will be performed with all {self.nGPUs} found GPUs.")
        else:
            self.log(f"Segmentation will be performed wtih all {self.nGPUs} found GPUs.")

        # set up threading
        if "threads" in self.config.keys():
            self.processes_per_GPU = self.config["threads"]
        else:
            self.processes_per_GPU = 1

        if self.nGPUs >= 1:
            self.n_processes = self.processes_per_GPU * self.nGPUs
        else:
            self.n_processes = self.processes_per_GPU

        # initialize a list of available GPUs
        gpu_id_list = []
        for gpu_ids in range(self.nGPUs):
            for _ in range(self.processes_per_GPU):
                gpu_id_list.append(gpu_ids)
        self.gpu_id_list = gpu_id_list

        self.log(
            f"GPU Status for segmentation is {self.use_GPU} with {self.nGPUs} GPUs found. Segmentation will be performed on the device {self.device} with {self.processes_per_GPU} processes per device in parallel."
        )

    def _check_filter_status(self):
        # check filter status in config
        if "filter_status" in self.config.keys():
            filter_status = self.config["filter_status"]
        else:
            filter_status = True  # always assumes that filtering is performed by default. Needs to be manually turned off if not desired.

        self.log(f"Filtering status for this segmentation is set to {filter_status}.")

        if not filter_status:
            # define path where the empty file should be generated
            filtered_path = os.path.join(self.directory, self.DEFAULT_FILTER_ADDTIONAL_FILE)

            with open(filtered_path, "w") as myfile:
                myfile.write("\n")

            self.log(
                f"Generated empty file at {filtered_path}. This marks that no filtering has been performed during segmentation and an additional step needs to be performed to populate this file with nucleus_id:cytosol_id matchings before running the extraction."
            )
        elif filter_status:
            self.log(
                "Filtering has been performed during segmentation. Nucleus and Cytosol IDs match. No additional steps are required."
            )

    def _load_input_image(self) -> np.ndarray:
        """Loads the input image from the sdatafile.

        Returns:
            np.ndarray: Input image as a np.ndarray.
        """
        start = timeit.default_timer()
        input_image = tempmmap.mmap_array_from_path(self._tmp_image_path)
        self.log(f"Time taken to load input image: {timeit.default_timer() - start}")
        return input_image

    def _select_relevant_channels(self, input_image):
        """transform image dtype and select segmentation channels

        The relevant channels for subsequent segmentation are determined by the channel ID's saved
        in `self.segmentation_channels`.

        Args:
            input_image (np.ndarray): Input image as a np.ndarray.

        Returns:
            np.ndarray: Transformed input image.
        """
        return input_image[self.segmentation_channels]

    def _transform_input_image(self, input_image):
        if isinstance(input_image, xarray.DataArray):
            input_image = input_image.data
        return input_image

    def _save_segmentation(self, labels: np.array, classes: list) -> None:
        """Helper function to save the results of a segmentation to file when generating a segmentation of a shard.

        Args:
            labels (np.array): Numpy array of shape ``(height, width)``. Labels are all data which are saved as integer values. These are mostly segmentation maps with integer values corresponding to the labels of cells.
            classes (list(int)): List of all classes in the labels array, which have passed the filtering step. All classes contained in this list will be extracted.

        """
        if self.deep_debug:
            self.log("saving segmentation")

        # size (C, H, W) is expected
        # dims are expanded in case (H, W) is passed

        labels = np.expand_dims(labels, axis=0) if len(labels.shape) == 2 else labels

        map_path = os.path.join(self.directory, self.DEFAULT_SEGMENTATION_FILE)
        hf = h5py.File(map_path, "w")

        # check if data container already exists and if so delete
        if self.DEFAULT_MASK_NAME in hf.keys():
            del hf[self.DEFAULT_MASK_NAME]
            self.log("labels dataset already existed in hdf5, dataset was deleted and will be overwritten.")

        hf.create_dataset(
            self.DEFAULT_MASK_NAME,
            data=labels,
            chunks=(1, self.chunk_size, self.chunk_size),
        )

        hf.close()

        # save classes
        self._check_filter_status()
        self._save_classes(classes)

        self.log("=== Finished segmentation of shard ===")

    def _save_segmentation_sdata_from_memmap(self, temp_file_path, masks=None):
        if masks is None:
            masks = ["nuclei", "cytosol"]

        # connect to the temp file as a dask array
        labels = dask_array_from_path(temp_file_path)

        if "nucleus" in masks:
            ix = masks.index("nucleus")

            self.filehandler._write_segmentation_sdata(labels[ix], self.nuc_seg_name, overwrite=self.overwrite)
            self.filehandler._add_centers(self.nuc_seg_name, overwrite=self.overwrite)

        if "cytosol" in masks:
            ix = masks.index("cytosol")
            self.filehandler._write_segmentation_sdata(labels[ix], self.cyto_seg_name, overwrite=self.overwrite)
            self.filehandler._add_centers(self.cyto_seg_name, overwrite=self.overwrite)

    def _save_segmentation_sdata(self, labels, classes, masks=None):
        if masks is None:
            masks = ["nucleus", "cytosol"]
        if self.is_shard:
            self._save_segmentation(labels, classes)
        else:
            if "nucleus" in masks:
                ix = masks.index("nucleus")

                if sc_any(labels[ix]):
                    self.filehandler._write_segmentation_sdata(labels[ix], self.nuc_seg_name, overwrite=self.overwrite)
                    self.filehandler._add_centers(self.nuc_seg_name, overwrite=self.overwrite)
                else:
                    self.log("No nuclei found in segmentation mask. Please check your processing")
                    warnings.warn("No nuclei found in segmentation mask. Please check your processing", stacklevel=2)

            if "cytosol" in masks:
                ix = masks.index("cytosol")

                if sc_any(labels[ix]):
                    self.filehandler._write_segmentation_sdata(labels[ix], self.cyto_seg_name, overwrite=self.overwrite)
                    self.filehandler._add_centers(self.cyto_seg_name, overwrite=self.overwrite)
                else:
                    self.log("No cytosols found in segmentation mask. Please check your processing")
                    warnings.warn("No cytosols found in segmentation mask. Please check your processing", stacklevel=2)

    def save_map(self, map_name):
        """Saves newly computed map.

        Args
            map_name (str): name of the map to be saved, as defined in ``self.maps``.

        Example:

            .. code-block:: python

                # declare all intermediate maps
                self.maps = {"myMap": None}

                # load intermediate maps if possible and get current processing step
                current_step = self.load_maps_from_disk()

                if current_step <= 0:
                    # do some computations

                    self.maps["myMap"] = myNumpyArray

                    # save map
                    self.save_map("myMap")
        """

        if self.maps[map_name] is None:
            self.log(f"Error saving map {map_name}, map is None")
        else:
            map_index = list(self.maps.keys()).index(map_name)

            # check if map contains more than one channel (3, 1024, 1024) vs (1024, 1024)
            if len(self.maps[map_name].shape) > 2:
                for i, channel in enumerate(self.maps[map_name]):
                    channel_name = f"{map_index}_{map_name}_{i}_map"
                    channel_path = os.path.join(self.directory, channel_name)

                    if self.debug and self.PRINT_MAPS_ON_DEBUG:
                        self.save_image(channel, save_name=channel_path)
            else:
                channel_name = f"{map_index}_{map_name}_map"
                channel_path = os.path.join(self.directory, channel_name)

                if self.debug and self.PRINT_MAPS_ON_DEBUG:
                    self.save_image(self.maps[map_name], save_name=channel_path)

    def save_image(self, array, save_name="", cmap="magma", **kwargs):
        if np.issubdtype(array.dtype.type, np.integer):
            self.log(f"{save_name} will be saved as tif")
            data = array.astype(np.uint16)
            im = Image.fromarray(data)
            im.save(f"{save_name}.tif")

        fig = plt.figure(frameon=False)
        fig.set_size_inches((10, 10))
        ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
        ax.set_axis_off()
        fig.add_axes(ax)
        ax.imshow(array, cmap=cmap, **kwargs)

        if save_name != "":
            plt.savefig(f"{save_name}.png")
            plt.show()
            plt.close()

    def get_output(self):
        return os.path.join(self.directory, self.DEFAULT_SEGMENTATION_FILE)

    def _initialize_as_shard(self, identifier, window, input_path, zarr_status=True):
        """Initialize Segmentation Step with further parameters needed for federated segmentation.

        Important:
            This function is intended for internal use by the :class:`ShardedSegmentation` helper class. In most cases it is not relevant to the creation of custom segmentation workflows.

        Args:
            identifier (int): Unique index of the shard.
            window (list(tuple)): Defines the window which is assigned to the shard. The window will be applied to the input. The first element refers to the first dimension of the image and so on. For example use ``[(0,1000),(0,2000)]`` To crop the image to `1000 px height` and `2000 px width` from the top left corner.
            input_path (str): Location of the input hdf5 file. During sharded segmentation the :class:`ShardedSegmentation` derived helper class will save the input image in form of a hdf5 file. This makes the input image available for parallel reading by the segmentation processes.
        """
        self.identifier = identifier
        self.window = window
        self.input_path = input_path
        self.save_zarr = zarr_status
        self.create_temp_dir()
        self.is_shard = True

    def _call_as_shard(self):
        """Wrapper function for calling a sharded segmentation.

        Important:
            This function is intended for internal use by the :class:`ShardedSegmentation` helper class. In most cases it is not relevant to the creation of custom segmentation workflows.
        """
        self.log(f"Beginning Segmentation of Shard with the slicing {self.window}")

        input_image = self._load_input_image()
        input_image = self._select_relevant_channels(input_image)

        # select the part of the image that is relevant for this shard
        input_image = input_image[
            :, self.window[0], self.window[1]
        ]  # for some segmentation workflows potentially only the first channel is required this is further selected down in that segmentation workflow

        if self.deep_debug:
            self.log(
                f"Input image of dtype {type(input_image)} with dimensions {input_image.shape} passed to sharded segmentation method."
            )

        if sc_any(input_image):
            try:
                self._execute_segmentation(input_image)
                self.clear_temp_dir()
            except (RuntimeError, ValueError, TypeError, IndexError) as e:
                print(
                    f"An error occurred when segmenting shard  with the tile id {self.directory} with the slicing {self.window}: {e}"
                )
                self.log(f"An error occurred: {e}")
                self.log(traceback.format_exc())
                self.clear_temp_dir()
        else:
            self.log(f"Shard in position [{self.window[0]}, {self.window[1]}] only contained zeroes.")
            try:
                super().__call_empty__(input_image)
                self.clear_temp_dir()
            except (RuntimeError, ValueError, TypeError, IndexError) as e:
                print(
                    f"An error occurred when segmenting shard  with the tile id {self.directory} with the slicing {self.window}: {e}"
                )
                self.log(f"An error occurred: {e}")
                self.log(traceback.format_exc())
                self.clear_temp_dir()

        self._clear_cache(vars_to_delete=["input_image"])

        # write out window location
        if self.deep_debug:
            self.log(f"Writing out window location to file at {self.directory}/window.csv")
        with open(f"{self.directory}/window.csv", "w") as f:
            f.write(f"{self.window}\n")

        self.log(f"Segmentation of Shard with the slicing {self.window} finished")

    def _save_classes(self, classes: list) -> None:
        """Helper function to save classes to a file when generating a segmentation of a shard."""
        # define path where classes should be saved
        filtered_path = os.path.join(self.directory, self.DEFAULT_CLASSES_FILE)

        to_write = "\n".join([str(i) for i in list(classes)])

        with open(filtered_path, "w") as myfile:
            myfile.write(to_write)

        self.log(f"Saved cell_id classes to file {filtered_path}.")

    def _save_benchmarking_times(
        self,
        image_size,
        transform_time,
        segmentation_time,
        total_time,
        max_shard_size=None,
        sharding_time=None,
        shard_resolving_time=None,
        time_per_shard=None,
    ):
        benchmarking_path = os.path.join(self.directory, self.DEFAULT_BENCHMARKING_FILE)

        benchmarking = pd.DataFrame(
            {
                "Size of the image": [image_size],
                "Number of GPUs used": [self.nGPUs],
                "Number of processes per GPU": [self.processes_per_GPU],
                "Total number of processes": [self.n_processes],
                "Shard max size": [max_shard_size if max_shard_size is not None else "N/A"],
                "Time taken for transformation": [transform_time],
                "Time taken for sharding": [sharding_time if sharding_time is not None else "N/A"],
                "Time taken for segmentation": [segmentation_time],
                "Time taken for shard resolving": [shard_resolving_time if shard_resolving_time is not None else "N/A"],
                "Time taken per shard": [time_per_shard if time_per_shard is not None else "N/A"],
                "Total time taken": [total_time],
            }
        )

        if os.path.exists(benchmarking_path):
            benchmarking.to_csv(benchmarking_path, mode="a", header=False, index=False)
        else:
            benchmarking.to_csv(benchmarking_path, index=False)

    def process(self, input_image):
        """Process the input image with the segmentation method."""
        image_size = input_image.shape
        input_image = self._select_relevant_channels(input_image)
        self._execute_segmentation(input_image)

        self._save_benchmarking_times(
            image_size=image_size,
            transform_time=self.transform_time,
            segmentation_time=self.segmentation_time,
            total_time=self.total_time,
        )


class ShardedSegmentation(Segmentation):
    """To perform a sharded segmentation where the input image is split into individual tiles (with overlap) that are processed idnividually before the results are joined back together."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not hasattr(self, "method"):
            raise AttributeError("No Segmentation method defined, please set attribute ``method``")

        # initialize a dummy instance of the segmentation method to determine which channels need to be loaded for segmentation
        test_method = self.method(
            self.config,
            directory=None,
            _tmp_image_path=None,
            nuc_seg_name=self.nuc_seg_name,
            cyto_seg_name=self.cyto_seg_name,
            filehandler=self.filehandler,
            project=None,
            project_location=None,
            debug=self.debug,
            overwrite=self.overwrite,
        )
        self.segmentation_channels = test_method.segmentation_channels

        if not hasattr(self, "_tmp_dir_path"):
            self.create_temp_dir()

    def _check_config(self):
        super()._check_config()

        # required config parameters for a sharded segmentation
        assert "shard_size" in self.config.keys(), "No shard size specified in config."
        assert "overlap_px" in self.config.keys(), "No overlap specified in config."
        assert "threads" in self.config.keys(), "No threads specified in config."

    def _calculate_sharding_plan(self, image_size) -> list:
        """Calculate the sharding plan for the given input image size."""

        _sharding_plan = []
        side_size = np.floor(np.sqrt(int(self.config["shard_size"])))
        shards_side = np.round(image_size / side_size).astype(int)
        shard_size = image_size // shards_side

        self.log(f"input image {image_size[0]} px by {image_size[1]} px")
        self.log(f"target_shard_size: {self.config['shard_size']}")
        self.log("sharding plan:")
        self.log(f"{shards_side[0]} rows by {shards_side[1]} columns")
        self.log(f"{shard_size[0]} px by {shard_size[1]} px")

        for y in range(shards_side[0]):
            for x in range(shards_side[1]):
                last_row = y == shards_side[0] - 1
                last_column = x == shards_side[1] - 1

                lower_y = y * shard_size[0]
                lower_x = x * shard_size[1]

                upper_y = (y + 1) * shard_size[0]
                upper_x = (x + 1) * shard_size[1]

                # add px overlap to each shard
                lower_y = lower_y - self.config["overlap_px"]
                lower_x = lower_x - self.config["overlap_px"]
                upper_y = upper_y + self.config["overlap_px"]
                upper_x = upper_x + self.config["overlap_px"]

                # make sure that each limit stays within the slides
                if lower_y < 0:
                    lower_y = 0
                if lower_x < 0:
                    lower_x = 0

                if last_row:
                    upper_y = image_size[0]

                if last_column:
                    upper_x = image_size[1]

                shard = (slice(lower_y, upper_y), slice(lower_x, upper_x))
                _sharding_plan.append(shard)

        return _sharding_plan

    def _get_sharding_plan(self, overwrite, force_read: bool = False) -> list:
        # check if a sharding plan already exists
        sharding_plan_path = f"{self.directory}/sharding_plan.csv"

        if os.path.isfile(sharding_plan_path):
            self.log(f"sharding plan already found in directory {sharding_plan_path}.")
            if overwrite:
                self.log("Overwriting existing sharding plan.")
                os.remove(sharding_plan_path)
            else:
                self.log("Reading existing sharding plan from file.")
                with open(sharding_plan_path) as f:
                    sharding_plan = [eval(line) for line in f.readlines()]
                    return sharding_plan

        if force_read:
            raise FileNotFoundError(
                "No sharding plan found in directory. Please run project.shard_segmentation() to generate a sharding plan. Can not complete a segmentation without the original sharding plan."
            )

        if self.config["shard_size"] >= np.prod(self.image_size):
            target_size = self.config["shard_size"]
            self.log(
                f"target size {target_size} is equal or larger to input image {np.prod(self.image_size)}. Sharding will not be used."
            )

            sharding_plan = [(slice(0, self.image_size[0]), slice(0, self.image_size[1]))]
        else:
            target_size = self.config["shard_size"]
            self.log(
                f"target size {target_size} is smaller than input image {np.prod(self.image_size)}. Sharding will be used."
            )
            sharding_plan = self._calculate_sharding_plan(self.image_size)

        # save sharding plan to file to be able to reload later
        self.log(f"Saving Sharding plan to file: {self.directory}/sharding_plan.csv")

        sharding_plan = list(enumerate(sharding_plan))
        with open(f"{self.directory}/sharding_plan.csv", "w") as f:
            for shard in sharding_plan:
                f.write(f"{shard}\n")

        return sharding_plan

    def _initialize_shard_list(self, sharding_plan):
        _shard_list = []

        self.input_path = self.filehandler.sdata_path

        for window in sharding_plan:
            i, window = window
            local_shard_directory = os.path.join(self.shard_directory, str(i))

            current_shard = self.method(
                self.config,
                directory=local_shard_directory,
                _tmp_image_path=self.input_image_path,
                nuc_seg_name=self.nuc_seg_name,
                cyto_seg_name=self.cyto_seg_name,
                filehandler=self.filehandler,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=self.overwrite,
                project=None,
            )

            current_shard._initialize_as_shard(i, window, self.input_path, zarr_status=False)
            _shard_list.append(current_shard)

        return _shard_list

    def _cleanup_shards(self, sharding_plan, keep_plots=False):
        file_identifiers_plots = [".png", ".tif", ".tiff", ".jpg", ".jpeg", ".pdf"]

        if keep_plots:
            self.log("Moving generated plots from shard directory to main directory.")
            for i, _window in enumerate(sharding_plan):
                local_shard_directory = os.path.join(self.shard_directory, str(i))
                for file in os.listdir(local_shard_directory):
                    if file.endswith(tuple(file_identifiers_plots)):
                        shutil.copyfile(
                            os.path.join(local_shard_directory, file),
                            os.path.join(self.directory, f"tile{i}_{file}"),
                        )
                        os.remove(os.path.join(local_shard_directory, file))

        # Add section here that cleans up the results from the tiles and deletes them to save memory
        self.log("Deleting intermediate tile results to free up storage space")
        shutil.rmtree(self.shard_directory, ignore_errors=True)

        self._clear_cache()

    def _resolve_sharding(self, sharding_plan):
        """
        The function iterates over a sharding plan and generates a new stitched hdf5 based segmentation.
        """

        self.log("resolve sharding plan")

        # ensure a temp directory is creates
        if not hasattr(self, "_tmp_dir_path"):
            self.create_temp_dir()

        label_size = (self.method.N_MASKS, self.image_size[0], self.image_size[1])

        # initialize an empty hdf5 file that will be filled with the results of the sharded segmentation
        # this is a workaround because as of yet labels can not be incrementally updated in spatialdata objects while being backed to disk

        hdf_labels_path = tempmmap.create_empty_mmap(
            shape=label_size,
            dtype=self.DEFAULT_SEGMENTATION_DTYPE,
            tmp_dir_abs_path=self._tmp_dir_path,
        )

        # clear temp directory used for sharding
        os.remove(self.input_image_path)
        self.log("Cleared temporary directory containing input image used for sharding.")

        hdf_labels = tempmmap.mmap_array_from_path(hdf_labels_path)

        class_id_shift = 0
        filtered_classes_combined = set()

        for window in sharding_plan:
            i, window = window
            timer = time.time()

            self.log(f"Stitching tile {i}")

            local_shard_directory = os.path.join(self.shard_directory, str(i))
            local_output = os.path.join(local_shard_directory, self.DEFAULT_SEGMENTATION_FILE)
            local_classes = os.path.join(local_shard_directory, "classes.csv")

            # check if this file exists otherwise abort process
            if not os.path.isfile(local_classes):
                sys.exit(
                    f"File {local_classes} does not exist. Processing of Shard {i} seems to be incomplete. \nAborting process to resolve sharding. Please run project.complete_segmentation() to regenerate this shard information."
                )

            # check to make sure windows match
            with open(f"{local_shard_directory}/window.csv") as f:
                window_local = eval(f.read())

            if window_local != window:
                Warning("Sharding plans do not match.")
                self.log("Sharding plans do not match.")
                self.log(f"Sharding plan found locally: {window_local}")
                self.log(f"Sharding plan found in sharding plan: {window}")
                self.log("Reading sharding window from local file and proceeding with that.")
                window = window_local

            local_hf = h5py.File(local_output, "r")
            local_hdf_labels = local_hf.get(self.DEFAULT_MASK_NAME)[:]

            shifted_map, edge_labels = shift_labels(
                local_hdf_labels,
                class_id_shift,
                return_shifted_labels=True,
                remove_edge_labels=True,
                dtype=self.DEFAULT_SEGMENTATION_DTYPE,
            )

            orig_input = hdf_labels[:, window[0], window[1]]
            _, x, y = orig_input.shape
            c_shifted, x_shifted, y_shifted = shifted_map.shape

            if x != x_shifted or y != y_shifted or c_shifted != self.method.N_MASKS:
                self.log(f"x: {x}, {x_shifted}")
                self.log(f"y: {y}, {y_shifted}")
                self.log(f"c: {self.method.N_MASKS}, {c_shifted}")

                Warning("Shapes do not match")
                self.log("Shapes do not match")
                self.log(f"window: {(window[0], window[1])}")
                self.log(f"shifted_map shape: {shifted_map.shape}")
                self.log(f"orig_input shape: {orig_input.shape}")

                raise ValueError("Shapes do not match. Please send this example to the developers for troubleshooting.")

            # since shards are computed with overlap there potentially already exist segmentations in the selected area that we wish to keep
            # if orig_input has a value that is not 0 (i.e. background) and the new map would replace this with 0 then we should keep the original value, in all other cases we should overwrite the values with the
            # new ones from the second shard

            # since segmentations resulting from cellpose are not necessarily deterministic we can not do this lookup on a pixel by pixel basis but need to edit
            # the segmentation mask to remove unwanted shapes before merging

            start_time_step1 = time.time()

            # assumptions: if a nucleus is discarded the cytosol must also be discarded and vice versa otherwise this could leave cells with only one of the two
            ids_discard = set(
                np.unique(orig_input[np.where((orig_input != 0) & (shifted_map != 0))])
            )  # gets all ids that potentially need to be discarded over both masks

            # we dont want to discard any id's' that touch the edge of the image
            # for these ids we can not ensure when processing this shard that the entire cell is present in the shard
            # if we would delete these ids we would potentially be deleting half of a cell
            edge_labels = set(_return_edge_labels(orig_input))
            ids_discard = ids_discard - edge_labels

            # set ids to 0 that we dont want to keep that are already in the final map
            orig_input[np.isin(orig_input, list(ids_discard))] = 0

            time_step1 = time.time() - start_time_step1

            start_time_step2 = time.time()

            # identify all ids from the new map that potentially could lead to problems
            problematic_ids = set(np.unique(shifted_map[np.where((orig_input != 0) & (shifted_map != 0))]))

            # remove all potentially problematic ids from the new map
            shifted_map[np.isin(shifted_map, list(problematic_ids))] = 0

            time_step2 = time.time() - start_time_step2

            if self.deep_debug:
                orig_input_manipulation = orig_input.copy()
                shifted_map_manipulation = shifted_map.copy()

                orig_input_manipulation[orig_input_manipulation > 0] = 1
                shifted_map_manipulation[shifted_map_manipulation > 0] = 2

                resulting_map = orig_input_manipulation + shifted_map_manipulation

                for _mask_ix in range(self.method.N_MASKS):
                    plt.figure()
                    plt.imshow(resulting_map[_mask_ix])
                    plt.title(
                        f"Combined segmentation mask {self.method.MASK_NAMES[_mask_ix]} after\n resolving sharding for region {i}"
                    )
                    plt.colorbar()
                    plt.show()
                    plt.savefig(
                        f"{self.directory}/combined_segmentation_mask_{self.method.MASK_NAMES[_mask_ix]}_{i}.png"
                    )

            start_time_step3 = time.time()
            shifted_map = np.where((orig_input != 0) & (shifted_map == 0), orig_input, shifted_map)

            time_step3 = time.time() - start_time_step3
            total_time = time_step1 + time_step2 + time_step3

            self.log(f"Time taken to cleanup overlapping shard regions for shard {i}: {total_time}s")

            # potential issue: this does not check if we create a cytosol without a matching nucleus? But this should have been implemented in altanas segmentation method
            # for other segmentation methods this could cause issues?? Potentially something to revisit in the future

            class_id_shift = np.max(shifted_map)  # get highest existing cell id and add it to the shift
            unique_ids = set(np.unique(shifted_map[0])[1:])  # get unique cellids in the shifted map

            # save results to hdf_labels
            hdf_labels[:, window[0], window[1]] = shifted_map

            # updated classes list
            filtered_classes_combined = (
                filtered_classes_combined - set(ids_discard)
            )  # ensure that the deleted ids are also removed from the classes list (class list is always in reference to the nucleus id!)
            filtered_classes_combined = filtered_classes_combined.union(
                unique_ids
            )  # get unique nucleus ids and add them to the combined filtered class

            if self.debug:
                self.log(f"Number of classes contained in shard after processing: {len(unique_ids)}")
                self.log(f"Number of Ids in filtered_classes after adding shard {i}: {len(filtered_classes_combined)}")

            # check if filtering of classes has been successfull
            # ideally the classes contained in the mask should be identical with those contained in the classes file
            # if this is not the case it is worth investigating and there it can be helpful to see which classes are contained in the mask but not in the classes file and vice versa
            if self.deep_debug:
                masks = hdf_labels[:, :, :]
                unique_ids = set(np.unique(masks[0])) - {0}
                self.log(f"Total number of classes in final segmentation after processing: {len(unique_ids)}")

                difference_classes = filtered_classes_combined.difference(unique_ids)
                self.log(
                    f"Classes contained in classes list that are not found in the segmentation mask: {difference_classes}"
                )

                difference_masks = unique_ids.difference(filtered_classes_combined)
                self.log(
                    f"Classes contained in segmentation mask that are not contained in classes list: {difference_masks}"
                )

                if len(difference_masks) > 0:
                    _masks = np.isin(masks[0], list(difference_masks))
                    plt.figure()
                    plt.imshow(_masks)
                    plt.set_title(
                        f"Classes in segmentation mask that are not contained in classes list after processing shard {i}"
                    )
                    plt.axis("off")
                    plt.show()

            local_hf.close()
            self.log(f"Finished stitching tile {i} in {time.time() - timer} seconds.")

            # remove background class
            filtered_classes_combined = filtered_classes_combined - {0}

            self.log(f"Number of filtered classes in Dataset: {len(filtered_classes_combined)}")

            # check filtering classes to ensure that segmentation run is properly tagged
            self._check_filter_status()

            # save newly generated class list
            self._save_classes(list(filtered_classes_combined))

        self.log("resolved sharding plan.")
        del hdf_labels  # ensure that memory is freed up

        # save final segmentation to sdata
        self._save_segmentation_sdata_from_memmap(hdf_labels_path, masks=self.method.MASK_NAMES)
        self.clear_temp_dir()  # ensure cleanup
        self.log("finished saving segmentation results to sdata object for sharded segmentation.")

        if not self.deep_debug:
            self._cleanup_shards(sharding_plan)

    def _initializer_function(self, gpu_id_list, n_processes):
        current_process().gpu_id_list = gpu_id_list
        current_process().n_processes = n_processes

    def _perform_segmentation(self, shard_list):
        # get GPU status
        self._setup_processing()

        # run in single-threaded mode
        if self.n_processes == 1:
            for shard in shard_list:
                shard._call_as_shard()

            self.log("Finished serial segmentation")

        # run in multithreaded mode
        elif self.n_processes > 1:
            with mp.get_context(self.context).Pool(
                processes=self.n_processes,
                initializer=self._initializer_function,
                initargs=[self.gpu_id_list, self.n_processes],
            ) as pool:
                list(
                    tqdm(
                        pool.imap(self.method._call_as_shard, shard_list),
                        total=len(shard_list),
                        desc="Segmenting Image Tiles",
                    )
                )
                pool.close()
                pool.join()
            self.log("Finished parallel segmentation")

    def process(self, input_image):
        """Process the input image with the sharded segmentation method.

        Important:
            This function is called automatically when a Segmentation Class is executed.


        Args:
            input_image (np.array): Input image to be processed. The input image should be a numpy array of shape (C, H, W) where C is the number of channels, H is the height of the image and W is the width of the image.

        """

        total_time_start = timeit.default_timer()

        start_transform = timeit.default_timer()

        # get proper level and shape of input image

        input_image = self._transform_input_image(input_image)
        input_image = self._select_relevant_channels(input_image)

        transform_time = timeit.default_timer() - start_transform

        self.input_image_path = self.filehandler._load_input_image_to_memmap(
            image=input_image, tmp_dir_abs_path=self._tmp_dir_path
        )
        self._clear_cache(vars_to_delete=[input_image])
        input_image = tempmmap.mmap_array_from_path(self.input_image_path)

        self.log("Mapped input image to memory-mapped array.")

        self.image_size = input_image.shape[1:]

        if self.deep_debug:
            self.log(
                f"Input image of dtype {type(input_image)} with dimensions {input_image.shape} passed to sharded segmentation method."
            )

        self.shard_directory = os.path.join(self.directory, self.DEFAULT_TILES_FOLDER)

        if not os.path.isdir(self.shard_directory):
            os.makedirs(self.shard_directory)
            self.log("Created new shard directory " + self.shard_directory)

        start_sharding = timeit.default_timer()

        # get sharding plan
        sharding_plan = self._get_sharding_plan(overwrite=self.overwrite)

        # generate shard list
        shard_list = self._initialize_shard_list(sharding_plan)
        self.log(
            f"sharding plan with {len(sharding_plan)} elements generated, sharding with {self.config['threads']} threads begins"
        )

        stop_sharding = timeit.default_timer()
        sharding_time = stop_sharding - start_sharding

        start_segmentation = timeit.default_timer()

        # perform segmentation
        self._perform_segmentation(shard_list)

        stop_segmentation = timeit.default_timer()
        segmentation_time = stop_segmentation - start_segmentation

        self._clear_cache(vars_to_delete=[shard_list])

        start_resolving = timeit.default_timer()

        self._resolve_sharding(sharding_plan)

        stop_resolving = timeit.default_timer()
        resolving_time = stop_resolving - start_resolving

        total_time_stop = timeit.default_timer()

        total_time = total_time_stop - total_time_start

        self.log(f"Total time taken for sharded segmentation: {total_time} seconds")

        # make sure to cleanup temp directories
        self.log("=== finished sharded segmentation === ")

        self._save_benchmarking_times(
            image_size=input_image.shape,
            transform_time=transform_time,
            segmentation_time=segmentation_time,
            total_time=total_time,
            max_shard_size=self.config["shard_size"],
            sharding_time=sharding_time,
            shard_resolving_time=resolving_time,
            time_per_shard=(total_time / len(sharding_plan)),
        )

    def complete_segmentation(self, input_image, force_run=False):
        """Complete an already started sharded segmentation of the provided input image.

        Args:
            input_image (np.array): Input image to be processed. The input image should be a numpy array of shape (C, H, W) where C is the number of channels, H is the height of the image and W is the width of the image.
            force_run (bool): If set to True the segmentation will be run even if a completed segmentation is already found in the sdata object. Default is False.
        """

        self.shard_directory = os.path.join(self.directory, self.DEFAULT_TILES_FOLDER)

        if "_tmp_dir_path" not in self.__dict__.keys():
            self.create_temp_dir()

        # check status of sdata object
        self.filehandler._check_sdata_status()

        if self.filehandler.nuc_seg_status:
            self.log("Nucleus segmentation already exists in sdata object.")
        if self.filehandler.cyto_seg_status:
            self.log("Cytosol segmentation already exists in sdata object.")

        # check to make sure that the shard directory exisits, if not exit and return error
        if not os.path.isdir(self.shard_directory):
            if not self.filehandler.nuc_seg_status or not self.filehandler.cyto_seg_status:
                raise FileNotFoundError(
                    "No shard directory found. Please run project.shard_segmentation() to generate a sharding instead of project.complete_segmentation()."
                )
            else:
                raise ValueError(
                    "Completed segmentation found and no shard directory found. Unclear what processing should be performed. If you believe this to be an error please send your example to the developers."
                )
        else:
            if not force_run:
                if self.filehandler.nuc_seg_status or self.filehandler.cyto_seg_status:
                    raise ValueError(
                        "Completed segmentation found in addition to shard directory. If you want to force overwrite of these segmentations with the sharded segmentation results found in the shard directory please rerurn the project.complete_segmentation() method with the force_run flag set to True."
                    )

        # check to see which tiles are incomplete
        tile_directories = os.listdir(self.shard_directory)
        incomplete_indexes = []
        for tile in tile_directories:
            if not os.path.isfile(f"{self.shard_directory}/{tile}/classes.csv"):
                incomplete_indexes.append(int(tile))
                self.log(f"Shard with ID {tile} not completed.")

        # get input image size
        input_image = self._transform_input_image(input_image)
        input_image = self._select_relevant_channels(input_image)

        self.input_image_path = self.filehandler._load_input_image_to_memmap(
            image=input_image, tmp_dir_abs_path=self._tmp_dir_path
        )
        self._clear_cache(vars_to_delete=[input_image])

        if len(incomplete_indexes) > 0:
            input_image = tempmmap.mmap_array_from_path(self.input_image_path)
            self.log("Mapped input image to memory-mapped array.")
        else:
            self.log("All shards already segmented. Proceeding to resolve sharding")

        self.image_size = input_image.shape[1:]

        # load sharding plan
        sharding_plan = self._get_sharding_plan(overwrite=False, force_read=True)

        # check to make sure that calculated sharding plan matches to existing sharding results
        assert (
            len(sharding_plan) == len(tile_directories)
        ), "Calculated a different number of shards than found shard directories. This indicates a mismatch between the current loaded config file and the config file used to generate the exisiting partial segmentation. Please rerun the complete segmentation to ensure accurate results."

        # select only those shards that did not complete successfully for further processing
        sharding_plan_complete = sharding_plan

        if len(incomplete_indexes) > 0:
            # adjust current sharding plan to only contain incomplete elements
            sharding_plan = [(i, shard) for i, shard in sharding_plan if i in incomplete_indexes]

            self.log(f"Adjusted sharding plan to only proceed with the {len(incomplete_indexes)} incomplete shards.")

            shard_list = self._initialize_shard_list(sharding_plan)
            self.log(
                f"sharding plan with {len(sharding_plan)} elements generated, sharding with {self.config['threads']} threads begins"
            )

            # perform segmentation
            self._perform_segmentation(shard_list)
            self._clear_cache(vars_to_delete=[shard_list])

        self._resolve_sharding(sharding_plan_complete)
        self._cleanup_shards(sharding_plan_complete, keep_plots=False)

        self.log("=== completed sharded segmentation === ")



"""Collection of functions to load pretrained models to use in the scPortrait environment."""

from pathlib import Path
from typing import Literal

import torch

from scportrait.data._dataloader import _download
from scportrait.tools.ml.plmodels import MultilabelSupervisedModel


def load_multilabelSupervised(
    checkpoint_path: str | Path,
    hparam_path: str | Path,
    model_type: str,
    eval: bool = True,
    device: Literal["cpu", "cuda", "mps"] = "cuda",
) -> MultilabelSupervisedModel:
    """Load a pretrained model uploaded to the github repository.

    Args:
        checkpoint_path: The path of the checkpoint file to load the pretrained model from
        hparam_path: The path of the hparams file containing the hyperparameters
        model_type: The type of the model, e.g., 'VGG1' or 'VGG2'
        eval: If True then the model will be returned in eval mode
        device: Device to load the model on, either "cuda" or "cpu"

    Returns:
        The pretrained multilabel classification model loaded from the checkpoint

    Examples:
        >>> model = load_multilabelSupervised("path/to/checkpoint.ckpt", "path/to/hparams.yaml", "resnet50")
        >>> print(model)
        MultilabelSupervisedModel(...)
    """
    model = MultilabelSupervisedModel.load_from_checkpoint(
        checkpoint_path, hparams_file=hparam_path, model_type=model_type, map_location=device
    )
    if eval:
        model.eval()
    return model


def get_data_dir() -> Path:
    """Get path to data that was packaged with scPortrait.

    Returns:
        Path to data directory
    """

    def find_root_by_folder(marker_folder: str, current_path: Path) -> Path | None:
        for parent in current_path.parents:
            if (parent / marker_folder).is_dir():
                return parent
        return None

    src_code_dir = find_root_by_folder("io", Path(__file__))

    if src_code_dir is None:
        raise FileNotFoundError("Could not find scPortrait source directory")
    data_dir = src_code_dir / "scportrait_data"

    return data_dir.absolute()


def autophagy_classifier(device: Literal["cuda", "cpu"] = "cuda") -> MultilabelSupervisedModel:
    """Load binary autophagy classification model from original SPARCS publication.

    Args:
        device: Device to load the model on, either "cuda" or "cpu"

    Returns:
        Pretrained autophagy classification model

    References:
        Schmacke NA, Mädler SC, Wallmann G, Metousis A, Bérouti M, Harz H,
        Leonhardt H, Mann M, Hornung V. SPARCS, a platform for genome-scale
        CRISPR screening for spatial cellular phenotypes. bioRxiv. 2023 Jun
        1;542416. doi: 10.1101/2023.06.01.542416.
    """
    data_dir = get_data_dir()
    save_path = data_dir / "vgg_autophagy_classifier"

    if not save_path.exists():
        _download(
            url="https://zenodo.org/records/13385705/files/vgg_autophagy_classifier.zip?download=1",
            output_path=str(data_dir),
            archive_format="zip",
        )

    checkpoint_path = save_path / "VGG2_autophagy_classifier2.1.cpkt"
    hparam_path = save_path / "hparams.yaml"

    # check if cuda is available
    if device == "cuda" and not torch.cuda.is_available():
        print("CUDA is not available. Loading model on CPU.")
        device = "cpu"

    model = load_multilabelSupervised(
        checkpoint_path=str(checkpoint_path), hparam_path=str(hparam_path), model_type="VGG2_old", device=device
    )
    return model



"""SpatialData file handling utilities for scPortrait."""

import os
import shutil
from pathlib import Path
from typing import Any, Literal, TypeAlias

import numpy as np
import xarray
from alphabase.io import tempmmap
from anndata import AnnData
from spatialdata import SpatialData
from spatialdata.models import Image2DModel, Labels2DModel, PointsModel, ShapesModel, TableModel
from spatialdata.transformations.transformations import Identity

from scportrait.pipeline._base import Logable
from scportrait.pipeline._utils.spatialdata_helper import (
    calculate_centroids,
    get_chunk_size,
)
from scportrait.tools.sdata.write import _write
from scportrait.tools.sdata.write._helper import add_element_sdata

ChunkSize2D: TypeAlias = tuple[int, int]
ChunkSize3D: TypeAlias = tuple[int, int, int]
ObjectType: TypeAlias = Literal["images", "labels", "points", "tables", "shapes"]

from scportrait.pipeline._utils.constants import (
    DEFAULT_CELL_ID_NAME,
    DEFAULT_CHUNK_SIZE_3D,
    DEFAULT_NAME_SINGLE_CELL_IMAGES,
)


class sdata_filehandler(Logable):
    def __init__(
        self,
        directory: str,
        sdata_path: str,
        input_image_name: str,
        nuc_seg_name: str,
        cyto_seg_name: str,
        centers_name: str,
        default_cell_id_name: str,
        debug: bool = False,
    ) -> None:
        """Initialize the SpatialData file handler.

        Args:
            directory: Base directory for operations
            sdata_path: Path to SpatialData file
            input_image_name: Name of input image in SpatialData
            nuc_seg_name: Name of nuclear segmentation
            cyto_seg_name: Name of cytoplasm segmentation
            centers_name: Name for cell centers
            debug: Enable debug mode
        """
        super().__init__(directory=directory, debug=debug)

        self.sdata_path = sdata_path
        self.input_image_name = input_image_name
        self.nuc_seg_name = nuc_seg_name
        self.cyto_seg_name = cyto_seg_name
        self.centers_name = centers_name
        self.default_cell_id_name = default_cell_id_name

    def _check_empty_sdata(self) -> bool:
        """Check if SpatialData object is empty.

        Returns:
            Whether SpatialData object is empty
        """
        empty_sdata_keys = {".zattrs", ".zgroup", "zmetadata"}
        if os.path.exists(self.sdata_path):
            keys = set(os.listdir(self.sdata_path))
            if keys == empty_sdata_keys:
                return True
            else:
                return False
        else:
            return True

    def _create_empty_sdata(self) -> SpatialData:
        """Create an empty SpatialData object.

        Returns:
            SpatialData object without any data
        """
        _sdata = SpatialData()
        _sdata.attrs["sdata_status"] = {
            "input_images": False,
            "nucleus_segmentation": False,
            "cytosol_segmentation": False,
            "centers": False,
        }
        return _sdata

    def _read_sdata(self) -> SpatialData:
        """Read or create SpatialData object.

        Returns:
            SpatialData object
        """
        if os.path.exists(self.sdata_path):
            _sdata = SpatialData.read(self.sdata_path)

        else:
            _sdata = self._create_empty_sdata()
            _sdata.write(self.sdata_path, overwrite=True)
        return _sdata

    def get_sdata(self) -> SpatialData:
        """Get the SpatialData object.

        Returns:
            SpatialData object
        """
        return self._read_sdata()

    def _force_delete_object(self, sdata: SpatialData, name: str, type: ObjectType) -> None:
        """Force delete an object from the SpatialData object and directory.

        Args:
            sdata: SpatialData object
            name: Name of object to delete
            type: Type of object ("images", "labels", "points", "tables", "shapes")
        """
        if name in sdata:
            del sdata[name]

        path = os.path.join(self.sdata_path, type, name)
        if os.path.exists(path):
            shutil.rmtree(path, ignore_errors=True)

    def _check_sdata_status(self, return_sdata: bool = False) -> SpatialData | None:
        """Check status of SpatialData objects.

        Args:
            return_sdata: Whether to return the SpatialData object

        Returns:
            SpatialData object if return_sdata is True, otherwise None
        """
        _sdata = self._read_sdata()

        self.input_image_status = self.input_image_name in _sdata.images
        self.nuc_seg_status = self.nuc_seg_name in _sdata.labels
        self.cyto_seg_status = self.cyto_seg_name in _sdata.labels
        self.nuc_centers_status = f"{self.centers_name}_{self.nuc_seg_name}" in _sdata.points
        self.cyto_centers_status = f"{self.centers_name}_{self.cyto_seg_name}" in _sdata.points

        updated_attrs = {
            "input_images": self.input_image_status,
            "nucleus_segmentation": self.nuc_seg_status,
            "cytosol_segmentation": self.cyto_seg_status,
            "nucleus_centers": self.nuc_centers_status,
            "cytosol_centers": self.cyto_centers_status,
        }

        # only update attrs on file if necessary to prevent potential data corruption due to no file space
        if "sdata_status" not in _sdata.attrs or updated_attrs != _sdata.attrs["sdata_status"]:
            _sdata.attrs["sdata_status"] = updated_attrs
            _sdata.write_metadata()  # ensure the metadata is updated on file
        if return_sdata:
            return _sdata
        return None

    def _get_input_image(self, sdata: SpatialData) -> xarray.DataArray:
        """Get input image from SpatialData object.

        Args:
            sdata: SpatialData object

        Returns:
            Input image as xarray DataArray

        Raises:
            ValueError: If input image not found
        """
        assert sdata.attrs["sdata_status"]["input_images"], "Input image not found in sdata object."
        if isinstance(sdata.images[self.input_image_name], xarray.DataTree):
            input_image = sdata.images[self.input_image_name]["scale0"].image
        elif isinstance(sdata.images[self.input_image_name], xarray.DataArray):
            input_image = sdata.images[self.input_image_name].image

        return input_image

    ## write elements to sdata object
    def _write_image_sdata(
        self,
        image,
        image_name: str,
        channel_names: list[str] = None,
        scale_factors: list[int] = None,
        chunks: ChunkSize3D = None,
        overwrite=False,
    ):
        """
        Write the supplied image to the spatialdata object.

        Args:
            image (dask.array): Image to be written to the spatialdata object.
            image_name (str): Name of the image to be written to the spatialdata object.
            channel_names list[str]: List of channel names for the image. Default is None.
            scale_factors list[int]: List of scale factors for the image. Default is [2, 4, 8]. This will load the image at 4 different resolutions to allow for fluid visualization.
            chunks (tuple): Chunk size for the image. Default is (1, 1000, 1000).
            overwrite (bool): Whether to overwrite existing data. Default is False.
        """
        _sdata = self._read_sdata()
        _write.image(
            _sdata,
            image,
            image_name,
            channel_names=channel_names,
            scale_factors=scale_factors,
            chunks=chunks,
            overwrite=overwrite,
        )
        self.log(f"Image {image_name} written to sdata object.")
        self._check_sdata_status()

    def _write_segmentation_sdata(
        self,
        segmentation: xarray.DataArray | np.ndarray,
        segmentation_label: str,
        chunks: ChunkSize2D = None,
        overwrite: bool = False,
    ) -> None:
        """Write segmentation data to SpatialData.

        Args:
            segmentation: Segmentation data to write
            segmentation_label: Label for the segmentation
            classes: Set of class names
            chunks: Chunk size for data storage
            overwrite: Whether to overwrite existing data
        """
        _write.labels(
            self._read_sdata(),
            segmentation,
            segmentation_label,
            chunks=chunks,
            overwrite=overwrite,
        )

        self.log(f"Segmentation {segmentation_label} written to sdata object.")
        self._check_sdata_status()

    def _write_points_object_sdata(self, points: PointsModel, points_name: str, overwrite: bool = False) -> None:
        """Write points object to SpatialData.

        Args:
            points: Points object to write
            points_name: Name for the points object
            overwrite: Whether to overwrite existing data
        """
        _sdata = self._read_sdata()
        add_element_sdata(_sdata, points, points_name, overwrite=overwrite)
        self.log(f"Points {points_name} written to sdata object.")
        self._check_sdata_status()

    def _write_table_sdata(
        self, adata: AnnData, table_name: str, segmentation_mask_name: str, overwrite: bool = False
    ) -> None:
        """Write anndata to SpatialData.

        Args:
            adata: AnnData object to write
            table_name: Name for the table object under which it should be saved
            segmentation_mask_name: Name of the segmentation mask that this table annotates
            overwrite: Whether to overwrite existing data

        Returns:
            None (writes to sdata object)
        """
        _sdata = self._read_sdata()

        assert isinstance(adata, AnnData), "Input data must be an AnnData object."
        assert segmentation_mask_name in _sdata.labels, "Segmentation mask not found in sdata object."

        # get obs and obs_indices
        obs = adata.obs
        if self.default_cell_id_name in obs.columns:
            obs_indices = obs[self.default_cell_id_name]
        else:
            raise ValueError(
                f"column `{self.default_cell_id_name}` not found in adata.obs object, but this object is expect to be able to annotate a segmentation."
            )

        # sanity checking
        assert len(obs_indices) == len(set(obs_indices)), "Instance IDs are not unique."
        cell_ids_mask = set(_sdata[f"{self.centers_name}_{segmentation_mask_name}"].index.values.compute())
        assert set(obs_indices).issubset(cell_ids_mask), "Instance IDs do not match segmentation mask cell IDs."

        obs["region"] = segmentation_mask_name
        obs["region"] = obs["region"].astype("category")

        adata.obs = obs
        table = TableModel.parse(
            adata,
            region=[segmentation_mask_name],
            region_key="region",
            instance_key=self.default_cell_id_name,
        )

        self._write_table_object_sdata(table, table_name, overwrite=overwrite)

    def _write_table_object_sdata(self, table: TableModel, table_name: str, overwrite: bool = False) -> None:
        """Write table object to SpatialData.

        Args:
            table: Table object to write
            table_name: Name for the table object
            overwrite: Whether to overwrite existing data
        """
        _sdata = self._read_sdata()
        add_element_sdata(_sdata, table, table_name, overwrite=overwrite)
        self.log(f"Table {table_name} written to sdata object.")
        self._check_sdata_status()

    def _write_shapes_object_sdata(self, shapes: ShapesModel, shapes_name: str, overwrite: bool = False) -> None:
        """Write shapes object to SpatialData.

        Args:
            shapes: Shapes object to write
            shapes_name: Name for the shapes object
            overwrite: Whether to overwrite existing data
        """
        _sdata = self._read_sdata()

        if overwrite:
            self._force_delete_object(_sdata, shapes_name, "shapes")

        _sdata.shapes[shapes_name] = shapes
        _sdata.write_element(shapes_name, overwrite=True)

        self.log(f"Shapes {shapes_name} written to sdata object.")

    def _get_centers(self, sdata: SpatialData, segmentation_label: str) -> PointsModel:
        """Get cell centers from segmentation.

        Args:
            sdata: SpatialData object
            segmentation_label: Label of segmentation to use

        Returns:
            Points model containing cell centers

        Raises:
            ValueError: If segmentation not found
        """
        if segmentation_label not in sdata.labels:
            raise ValueError(f"Segmentation {segmentation_label} not found in sdata object.")

        mask = sdata.labels[segmentation_label]
        if isinstance(mask, xarray.DataTree):
            mask = mask.scale0.image
        centers = calculate_centroids(mask)
        return centers

    def _add_centers(self, segmentation_label: str, overwrite: bool = False) -> None:
        """Add cell centers from segmentation.

        Args:
            segmentation_label: Label of segmentation to use
            overwrite: Whether to overwrite existing centers
        """
        _sdata = self._read_sdata()
        centroids_object = self._get_centers(_sdata, segmentation_label)
        centers_name = f"{self.centers_name}_{segmentation_label}"
        self._write_points_object_sdata(centroids_object, centers_name, overwrite=overwrite)

    ## load elements from sdata to a memory mapped array
    def _load_input_image_to_memmap(
        self, tmp_dir_abs_path: str | Path, image: np.typing.NDArray[Any] | None = None
    ) -> str:
        """Helper function to load the input image from sdata to memory mapped temp arrays for faster access.

        Loading happens in a chunked manner to avoid memory issues.

        Args:
            tmp_dir_abs_path: Absolute path to the directory where the memory mapped arrays should be stored.
            image: Optional pre-loaded image array to process.

        Returns:
            Path to the memory mapped array. Can be reconneted to using the `mmap_array_from_path`
            function from the alphabase.io.tempmmap module.

        Raises:
            ValueError: If input image is not found in sdata object.
        """
        if image is None:
            _sdata = self._check_sdata_status(return_sdata=True)

            if not self.input_image_status:
                raise ValueError("Input image not found in sdata object.")

            image = self._get_input_image(_sdata)
        shape = image.shape

        # initialize empty memory mapped arrays to store the data
        path_input_image = tempmmap.create_empty_mmap(shape=shape, dtype=image.dtype, tmp_dir_abs_path=tmp_dir_abs_path)
        input_image_mmap = tempmmap.mmap_array_from_path(path_input_image)

        Z: int | None = None
        if len(shape) == 3:
            C, Y, X = shape

        elif len(shape) == 4:
            Z, C, Y, X = shape

        if Z is not None:
            for z in range(Z):
                for c in range(C):
                    input_image_mmap[z][c] = image[z][c].compute()
        else:
            for c in range(C):
                input_image_mmap[c] = image[c].compute()

        # cleanup the cache
        del input_image_mmap, image

        return path_input_image

    def _load_seg_to_memmap(
        self,
        seg_name: list[str],
        tmp_dir_abs_path: str | Path,
    ) -> str:
        """Helper function to load segmentation masks from sdata to memory mapped temp arrays for faster access.

        Loading happens in a chunked manner to avoid memory issues.

        Args:
            seg_name: List of segmentation element names that should be loaded found in the sdata object.
                The segmentation elments need to have the same size.
            tmp_dir_abs_path: Absolute path to the directory where the memory mapped arrays should be stored.

        Returns:
            Path to the memory mapped array. Can be reconneted to using the `mmap_array_from_path`
            function from the alphabase.io.tempmmap module.

        Raises:
            AssertionError: If not all segmentation elements are found in sdata object or if shapes don't match.
        """
        _sdata = self._check_sdata_status(return_sdata=True)

        assert all(
            seg in _sdata.labels for seg in seg_name
        ), "Not all passed segmentation elements found in sdata object."

        seg_objects = [_sdata.labels[seg] for seg in seg_name]

        # ensure we get the correct level of segmentation
        for i, seg in enumerate(seg_objects):
            if isinstance(seg, xarray.DataTree):
                seg = seg.scale0.image
                seg_objects[i] = seg

        shapes = [seg.shape for seg in seg_objects]

        Z: int | None = None
        Y: int | None = None
        X: int | None = None
        for shape in shapes:
            if len(shape) == 2:
                if Y is None:
                    Y, X = shape
                else:
                    assert Y == shape[0]
                    assert X == shape[1]
            elif len(shape) == 3:
                if Z is None:
                    Z, Y, X = shape
                else:
                    assert Z == shape[0]
                    assert Y == shape[1]
                    assert X == shape[2]

        n_masks = len(seg_objects)

        if Z is not None and Y is not None and X is not None:
            shape = (n_masks, Z, Y, X)
        elif Y is not None and X is not None:
            shape = (n_masks, Y, X)
        else:
            raise ValueError("Unable to determine shape from segmentation masks")

        # initialize empty memory mapped arrays to store the data
        path_seg_masks = tempmmap.create_empty_mmap(
            shape=shape, dtype=seg_objects[0].data.dtype, tmp_dir_abs_path=tmp_dir_abs_path
        )

        seg_masks = tempmmap.mmap_array_from_path(path_seg_masks)

        for i, seg in enumerate(seg_objects):
            if Z is not None:
                for z in range(Z):
                    seg_masks[i][z] = seg.data[z].compute()
            else:
                seg_masks[i] = seg.data.compute()

        # cleanup the cache
        del seg_masks, seg_objects, seg

        return path_seg_masks



from collections.abc import Sequence
from math import floor

import numpy as np
import pandas as pd
import torch
from scipy.sparse import csr_matrix
from torch.utils.data import ConcatDataset, Dataset


def combine_datasets_balanced(
    list_of_datasets: list[Dataset],
    class_labels: list[str | int],
    train_per_class: int,
    val_per_class: int,
    test_per_class: int,
    seed: int | None = None,
    return_residual: bool = False,
    concat_datasets: bool = True,
) -> (
    tuple[ConcatDataset, ConcatDataset, ConcatDataset]
    | tuple[ConcatDataset, ConcatDataset, ConcatDataset, ConcatDataset]
    | tuple[list[Dataset], list[Dataset], list[Dataset]]
    | tuple[list[Dataset], list[Dataset], list[Dataset], list[Dataset]]
):
    """Combine multiple datasets to create a balanced dataset for train, val, and test sets.

    Args:
        list_of_datasets: List of datasets to be combined
        class_labels: List of class labels present in the datasets
        train_per_class: Number of samples per class in the train set
        val_per_class: Number of samples per class in the validation set
        test_per_class: Number of samples per class in the test set
        seed: Random seed for reproducibility
        return_residual: boolean value indicating of left over elements not allocated to one of the datasets should be returned as a seperate dataset
        concat_datasets: if the returned datasets should be concatenated or returned as a list

    Returns:
        Tuple containing:
            - train dataset(s) with balanced samples per class
            - validation dataset(s) with balanced samples per class
            - test dataset(s) with balanced samples per class
            - potentially: residual dataset(s) with all residual samples (unbalanced)

    Raises:
        ValueError: If dataset is too small to be split into requested sizes
    """
    elements = [len(el) for el in list_of_datasets]
    rows = np.arange(len(list_of_datasets))

    # Map labels to 0-based consecutive integers
    unique_labels = sorted(set(class_labels))
    label_to_index = {label: idx for idx, label in enumerate(unique_labels)}
    indexed_labels = [label_to_index[label] for label in class_labels]

    # create dataset fraction array of len(list_of_datasets)
    # mat = csr_matrix((elements, (rows, class_labels))).toarray()
    mat = csr_matrix((elements, (rows, indexed_labels))).toarray()
    cells_per_class = np.sum(mat, axis=0)
    normalized = mat / cells_per_class
    dataset_fraction = np.sum(normalized, axis=1)

    train_dataset: list[Dataset] = []
    test_dataset: list[Dataset] = []
    val_dataset: list[Dataset] = []
    residual_dataset: list[Dataset] = []

    if np.sum(pd.Series(class_labels).value_counts() > 1) == 0:
        for dataset, label, fraction in zip(list_of_datasets, class_labels, dataset_fraction, strict=False):
            print(dataset, label, fraction)

            train_size = floor(train_per_class)
            test_size = floor(test_per_class)
            val_size = floor(val_per_class)

            residual_size = len(dataset) - train_size - test_size - val_size

            if residual_size < 0:
                raise ValueError(
                    f"Dataset with length {len(dataset)} is too small to be split into "
                    f"test set of size {test_size} and train set of size {train_size} "
                    f"and validation set of size {val_size}. Use smaller sizes."
                )

            if seed is not None:
                print(f"Using seeded generator with seed {seed} to split dataset")
                gen = torch.Generator()
                gen.manual_seed(seed)
                splits = torch.utils.data.random_split(
                    dataset, [train_size, test_size, val_size, residual_size], generator=gen
                )
            else:
                splits = torch.utils.data.random_split(dataset, [train_size, test_size, val_size, residual_size])
            train, test, val, residual = splits
            train_dataset.append(train)
            test_dataset.append(test)
            val_dataset.append(val)
            if return_residual:
                residual_dataset.append(residual)
    else:
        for dataset, fraction in zip(list_of_datasets, dataset_fraction, strict=False):
            train_size = int(np.round(train_per_class * fraction))
            test_size = int(np.round(test_per_class * fraction))
            val_size = int(np.round(val_per_class * fraction))

            residual_size = len(dataset) - train_size - test_size - val_size

            if residual_size < 0:
                raise ValueError(
                    f"Dataset with length {len(dataset)} is too small to be split into "
                    f"test set of size {test_size}, train set of size {train_size}, "
                    f"and validation set of size {val_size}. Use smaller sizes."
                )

            if seed is not None:
                print(f"Using seeded generator with seed {seed} to split dataset")
                gen = torch.Generator()
                gen.manual_seed(seed)
                splits = torch.utils.data.random_split(
                    dataset, [train_size, test_size, val_size, residual_size], generator=gen
                )
            else:
                splits = torch.utils.data.random_split(dataset, [train_size, test_size, val_size, residual_size])

            train, test, val, residual = splits
            train_dataset.append(train)
            test_dataset.append(test)
            val_dataset.append(val)

            if return_residual:
                residual_dataset.append(residual)

    if concat_datasets:
        train_dataset = ConcatDataset(train_dataset)
        val_dataset = ConcatDataset(val_dataset)
        test_dataset = ConcatDataset(test_dataset)
        if return_residual:
            residual_dataset = ConcatDataset(residual_dataset)

    if return_residual:
        return (train_dataset, val_dataset, test_dataset, residual_dataset)
    else:
        return (train_dataset, val_dataset, test_dataset)


def split_dataset_fractions(
    list_of_datasets: list[Dataset],
    train_size: int | None = None,
    test_size: int | None = None,
    val_size: int | None = None,
    fractions: Sequence[float] | None = None,
    seed: int | None = None,
) -> tuple[ConcatDataset, ConcatDataset, ConcatDataset]:
    """Split datasets into train, test, and validation sets based on sizes or fractions.

    Args:
        list_of_datasets: List of datasets to split
        train_size: Number of samples in the train set
        test_size: Number of samples in the test set
        val_size: Number of samples in the validation set
        fractions: List of fractions for train/test/val split, must sum to 1
        seed: Random seed for reproducibility

    Returns:
        Tuple containing:
            - Train dataset
            - Test dataset
            - Validation dataset

    Raises:
        ValueError: If fractions don't sum to 1 or if dataset is too small for requested sizes
    """
    train_dataset: list[Dataset] = []
    test_dataset: list[Dataset] = []
    val_dataset: list[Dataset] = []

    for dataset in list_of_datasets:
        if fractions is not None:
            if sum(fractions) != 1:
                raise ValueError("Provided fractions should sum to 1")

            if seed is not None:
                gen = torch.Generator()
                gen.manual_seed(seed)
                splits = torch.utils.data.random_split(dataset, fractions, generator=gen)
            else:
                splits = torch.utils.data.random_split(dataset, fractions)

            train, test, val = splits
            print(
                f"Dataset {list_of_datasets.index(dataset)}:\n"
                f"Train: {len(train)}\n"
                f"Test: {len(test)}\n"
                f"Validation: {len(val)}"
            )

            train_dataset.append(train)
            test_dataset.append(test)
            val_dataset.append(val)
        else:
            residual_size = len(dataset) - train_size - test_size - val_size
            if residual_size < 0:
                raise ValueError(f"Dataset with length {len(dataset)} is too small to be split into requested sizes")

            if seed is not None:
                gen = torch.Generator()
                gen.manual_seed(seed)
                splits = torch.utils.data.random_split(
                    dataset, [train_size, test_size, val_size, residual_size], generator=gen
                )
            else:
                splits = torch.utils.data.random_split(dataset, [train_size, test_size, val_size, residual_size])

            train, test, val, _ = splits
            print(
                f"Dataset {list_of_datasets.index(dataset)}:\n"
                f"Train: {len(train)}\n"
                f"Test: {len(test)}\n"
                f"Validation: {len(val)}"
            )

            train_dataset.append(train)
            test_dataset.append(test)
            val_dataset.append(val)

    # Convert to ConcatDataset objects
    train_concat = ConcatDataset(train_dataset)
    test_concat = ConcatDataset(test_dataset)
    val_concat = ConcatDataset(val_dataset)

    print(
        f"Total sizes:\n" f"Train: {len(train_concat)}\n" f"Test: {len(test_concat)}\n" f"Validation: {len(val_concat)}"
    )

    return train_concat, test_concat, val_concat



import numpy as np
from scipy.ndimage import fourier_shift
from skimage.registration import phase_cross_correlation


def apply_shift(target_img: np.ndarray, shift: np.ndarray) -> np.ndarray:
    """Applies a Fourier shift to the target image.

    This function shifts the input image in the Fourier domain by the specified amount
    and returns the corrected image. This image is usually another staining from another cycle.

    Args:
        target_img: The input image to be shifted.
        shift: The amount of shift to apply.

    Returns:
        corrected_image: The shifted image.
    """
    corrected_image = fourier_shift(np.fft.fftn(target_img), np.array(shift))
    corrected_image = np.fft.ifftn(corrected_image).real  # Take real part
    return corrected_image


def get_aligned_img(
    source_img: np.ndarray, target_img: np.ndarray, normalize_images: bool = True
) -> tuple[np.ndarray, np.ndarray]:
    """Registers the target image to the source image using phase cross-correlation.

    This function calculates the shift required to align the target image with the source image
    using phase cross-correlation and applies that shift. The shift can also be used for
    aligning other stainings of the same cycle as the target image.

    Based on:
    https://scikit-image.org/docs/0.23.x/auto_examples/registration/plot_register_translation.html

    Args:
        source_img: The reference image to which the target image will be aligned.
            Usually the DAPI staining of a cycle.
        target_img: The image to be registered to the source image.
            Usually the DAPI staining of another cycle.
        normalize_images: If True, normalize images to the range [0, 1] before registration.
            Defaults to True.

    Returns:
        A tuple containing:
            corrected_image: The registered (shifted) target image.
            shift: The shift vector that was applied to align the target image.
    """
    if normalize_images:
        source_img = (source_img - source_img.min()) / (source_img.max() - source_img.min())
        target_img = (target_img - target_img.min()) / (target_img.max() - target_img.min())
    shift, error, diffphase = phase_cross_correlation(source_img, target_img, upsample_factor=100)
    corrected_image = apply_shift(target_img, shift)  # Apply shift
    return corrected_image, shift
